"""
Create torrent file
"""

import os
import queue

from .. import errors
from ..utils import daemon, fs, torrent
from . import base

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CreateTorrentJob(base.JobBase):
    """
    Create torrent file

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``announce_url``
            Emitted before and after successful announce URL retrieval.
            Registered callbacks get an :class:`Ellipsis` to indicate the
            retrieval attempt and the announce URL if the attempt was
            successful.

        ``file_tree``
            Emitted when torrent creation begins. Registered callbacks get
            nested `(file_name, file_size)` tuples (see
            :func:`~.utils.torrent.create`) as a positional argument. (See also
            :func:`.fs.format_file_tree`.)

        ``file_list``
            Emitted when torrent creation begins. Registered callbacks get a
            :class:`list` of paths to existing files that are in the torrent as
            a positional argument.

        ``progress_update``
            Emitted at roughly equal intervals to provide information about the
            torrent creation progress. Registered callbacks get a
            :class:`~.torrent.CreateTorrentProgress` or
            :class:`~.torrent.FindTorrentProgress` instance as a positional
            argument.
    """

    name = 'torrent'
    label = 'Torrent'
    cache_id = None

    def initialize(self, *, tracker, content_path, exclude_files=(), reuse_torrent_path=None):
        """
        Set internal state

        :param TrackerBase tracker: Return value of :func:`.trackers.tracker`
        :param content_path: Path to file or directory
        :param reuse_torrent_path: Path to existing torrent file to reuse piece
            hashes from (see :func:`~.utils.torrent.create`)
        :param exclude_files: Sequence of glob patterns (:class:`str`) and
            :class:`re.Pattern` objects (return value from :func:`re.compile`)
            that are matched against the relative path within the generated
            torrent

            Glob patterns are matched case-insensitively.

            .. note:: This sequence is combined with
                      :attr:`.TrackerBase.options`\\ ``["exclude"]``.
        """
        self._tracker = tracker
        self._content_path = content_path
        self._reuse_torrent_path = reuse_torrent_path
        self._torrent_path = os.path.join(
            self.home_directory,
            f'{fs.basename(content_path)}.{tracker.name.lower()}.torrent',
        )

        self._exclude_files = list(self._tracker.options['exclude'])
        self._exclude_files.extend(exclude_files)

        self.signal.add('announce_url')
        self.signal.add('file_tree')
        self.signal.add('file_list')
        self.signal.add('progress_update')
        self._torrent_process = None
        self._activity = ''

    @property
    def activity(self):
        """
        What is currently being done

        ``"hashing"``
            If we are hashing files.

        ``"searching"``
            If we are searching for a torrent we can copy hashes from.

        ``"verifying"``
            If we are hash-checking a potentially matching torrent.

        ``""``
            None of the above is happening (e.g. we are getting the file tree).
        """
        return self._activity

    def execute(self):
        """Get announce URL from `tracker`, then execute torrent creation subprocess"""
        self._get_announce_url_task = self.attach_task(self._get_announce_url())
        self._get_announce_url_task.add_done_callback(
            lambda task: self._start_torrent_creation_process(task.result())
        )

    async def _get_announce_url(self):
        self._activity = 'announce_url'
        self.signal.emit('announce_url', Ellipsis)
        try:
            announce_url = await self._tracker.get_announce_url()
        except errors.RequestError as e:
            self.error(e)
        else:
            self.signal.emit('announce_url', announce_url)
            return announce_url
        finally:
            self._activity = ''

    def _start_torrent_creation_process(self, announce_url):
        self._torrent_process = daemon.DaemonProcess(
            name=self.name,
            target=_torrent_process,
            kwargs={
                'content_path': self._content_path,
                'announce': announce_url,
                'source': self._tracker.options['source'],
                'torrent_path': self._torrent_path,
                'exclude': self._exclude_files,
                'use_cache': not self.ignore_cache,
                'reuse_torrent_path': self._reuse_torrent_path,
                'randomize_infohash': self._tracker.options['randomize_infohash'],
                'piece_size_calculator': self._tracker.calculate_piece_size,
                'piece_size_min_max_calculator': self._tracker.calculate_piece_size_min_max,
            },
            init_callback=self._handle_files,
            info_callback=self._handle_info_update,
            error_callback=self._handle_error,
            result_callback=self._handle_torrent_created,
            finished_callback=self.finish,
        )
        self._torrent_process.start()

    def finish(self):
        """Terminate torrent creation subprocess and finish"""
        if self._torrent_process:
            self._torrent_process.stop()
        super().finish()

    def cancel_search(self):
        """
        Stop searching for a torrent we can copy hashes from

        If we are not currently searching for a torrent, do nothing.
        """
        if self._activity == 'searching' and self._torrent_process:
            self._torrent_process.send(daemon.MsgType.terminate, torrent.SKIP_SEARCHING)

    def _handle_files(self, files):
        self.signal.emit('file_tree', files.tree)
        self.signal.emit('file_list', files.list)

    def _handle_info_update(self, info):
        if isinstance(info, torrent.CreateTorrentProgress):
            self._activity = 'hashing'
            self.signal.emit('progress_update', info)

        elif isinstance(info, torrent.FindTorrentProgress):
            if info.status == 'verifying':
                self._activity = 'verifying'
            elif info.status == 'hit':
                _log.debug('Reusing torrent: %r', info.filepath)
                self._activity = ''
            else:
                self._activity = 'searching'
            self.signal.emit('progress_update', info)

        else:
            self._activity = ''
            raise RuntimeError('Unexpected info update: {info!r}')

    def _handle_torrent_created(self, torrent_path=None):
        _log.debug('Torrent created: %r', torrent_path)
        if torrent_path:
            self.send(torrent_path)

    def _handle_error(self, error):
        if isinstance(error, BaseException):
            _log.debug('CreateTorrentJob: Raising %r', error)
            self.exception(error)
        else:
            _log.debug('CreateTorrentJob: Reporting %r', error)
            self.error(error)


def _torrent_process(output_queue, input_queue, *args, **kwargs):
    def read_input_queue():
        try:
            typ, msg = input_queue.get_nowait()
            return typ, msg
        except queue.Empty:
            return (None, None)

    def init_callback(files):
        typ, msg = read_input_queue()
        if typ == daemon.MsgType.terminate:
            return 'cancel'
        else:
            output_queue.put((daemon.MsgType.init, files))

    def progress_callback(progress):
        typ, msg = read_input_queue()
        if typ == daemon.MsgType.terminate:
            if (
                    # We were told to skip searching (not to cancel torrent creation altogether)
                    msg == torrent.SKIP_SEARCHING
                    # We are searching
                    and isinstance(progress, torrent.FindTorrentProgress)
                    # No match found yet (False = no match, None = Verifying potential match)
                    and progress.status is not True
            ):
                # Stop searching for reusable torrent
                return torrent.SKIP_SEARCHING
            else:
                # Cancel torrent creation altogether
                return 'cancel'
        else:
            output_queue.put((daemon.MsgType.info, progress))

    kwargs['init_callback'] = init_callback
    kwargs['progress_callback'] = progress_callback
    try:
        torrent_path = torrent.create(*args, **kwargs)
    except errors.TorrentCreateError as e:
        output_queue.put((daemon.MsgType.error, str(e)))
    else:
        output_queue.put((daemon.MsgType.result, torrent_path))


class AddTorrentJob(base.QueueJobBase):
    """
    Add torrent(s) to a BitTorrent client

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``adding``
            Emitted when attempting to add a torrent. Registered callbacks get
            the path to the torrent file as a positional argument.

        ``added``
            Emitted when the torrent was added successfully. Registered
            callbacks get the added torrent's info hash as a positional
            argument.
    """

    name = 'add-torrent'
    label = 'Add Torrent'
    cache_id = None  # Don't cache output

    def initialize(self, *, btclient, enqueue=()):
        """
        Set internal state

        :param btclient: :class:`~.btclient.BtClient` instance
        :param enqueue: Sequence of torrent file paths to add

        If `enqueue` is given and not empty, this job is finished as soon as its
        last item is added.
        """
        self._btclient = btclient
        self.signal.add('adding')
        self.signal.add('added')
        self.signal.register('adding', lambda tp: setattr(self, 'info', f'Adding {fs.basename(tp)}'))
        self.signal.register('added', lambda _: setattr(self, 'info', ''))
        self.signal.register('finished', lambda _: setattr(self, 'info', ''))
        self.signal.register('error', lambda _: setattr(self, 'info', ''))

    async def handle_input(self, torrent_path):
        _log.debug('Adding %s to %s', torrent_path, self._btclient.name)
        self.signal.emit('adding', torrent_path)

        try:
            infohash = await self._btclient.add_torrent(torrent_path)
        except errors.TorrentAddError as e:
            self.error(e)
        else:
            self.send(infohash)
            self.signal.emit('added', infohash)


class CopyTorrentJob(base.QueueJobBase):
    """
    Copy file(s)

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``copying``
            Emitted when attempting to copy a file. Registered callbacks get the
            source file path as a positional argument.

        ``copied``
            Emitted when the copy attempt ended. Registered callbacks get the
            destination file path (success) or the source file path (failure) as
            a positional argument.
    """

    name = 'copy-torrent'
    label = 'Copy Torrent'
    cache_id = None  # Don't cache output

    def initialize(self, *, destination, enqueue=()):
        """
        Set internal state

        :param destination: Where to put the torrent(s)
        :param enqueue: Sequence of file paths to copy

        If `enqueue` is given and not empty, this job is finished as soon as its
        last item is copied.
        """
        self._destination = None if not destination else str(destination)
        self.signal.add('copying')
        self.signal.add('copied')
        self.signal.register('copying', lambda fp: setattr(self, 'info', f'Copying {fs.basename(fp)}'))
        self.signal.register('copied', lambda _: setattr(self, 'info', ''))
        self.signal.register('finished', lambda _: setattr(self, 'info', ''))
        self.signal.register('error', lambda _: setattr(self, 'info', ''))

    MAX_FILE_SIZE = 10 * 2**20  # 10 MiB
    """Upper limit of acceptable file size"""

    async def handle_input(self, filepath):
        _log.debug('Copying %s to %s', filepath, self._destination)

        if not os.path.exists(filepath):
            self.error(f'{filepath}: No such file')

        elif os.path.getsize(filepath) > self.MAX_FILE_SIZE:
            self.error(f'{filepath}: File is too large')

        else:
            self.signal.emit('copying', filepath)

            import shutil
            try:
                new_path = shutil.copy2(filepath, self._destination)
            except OSError as e:
                msg = e.strerror if e.strerror else str(e)
                self.error(f'Failed to copy {filepath} to {self._destination}: {msg}')
                # Default to original torrent path
                self.send(filepath)
            else:
                self.send(new_path)
                self.signal.emit('copied', new_path)



"""
Concrete :class:`~.base.TrackerConfigBase` subclass for BHD
"""

import base64

from ...utils import argtypes, configfiles, imghosts, types
from ..base import TrackerConfigBase


class BhdTrackerConfig(TrackerConfigBase):
    defaults = {
        'upload_url': base64.b64decode('aHR0cHM6Ly9iZXlvbmQtaGQubWUvYXBpL3VwbG9hZA==').decode('ascii'),
        'announce_url': configfiles.config_value(
            value=base64.b64decode('aHR0cHM6Ly9iZXlvbmQtaGQubWUvYW5ub3VuY2U=').decode('ascii'),
            description='The announce URL without the private passkey.',
        ),
        'announce_passkey': configfiles.config_value(
            value='',
            description=(
                'The private part of the announce URL.\n'
                'Get it from the website: My Security -> Passkey'
            ),
        ),
        'apikey': configfiles.config_value(
            value='',
            description=(
                'Your personal private API key.\n'
                'Get it from the website: My Security -> API key'
            ),
        ),
        'source': 'BHD',
        'anonymous': configfiles.config_value(
            value=types.Bool('no'),
            description='Whether your username is displayed on your uploads.',
        ),
        'draft': configfiles.config_value(
            value=types.Bool('no'),
            description=(
                'Whether your uploads are stashed under Torrents -> Drafts '
                'after the upload instead of going live.'
            ),
        ),
        'image_host': types.Choice('imgbox', options=imghosts.imghost_names()),
        'screenshots': configfiles.config_value(
            value=types.Integer(4, min=3, max=10),
            description='How many screenshots to make.',
        ),
        'exclude': (
            r'\.(?i:txt|jpg|jpeg|png|sfv|md5)$',
            r'/(?i:sample)',
        ),
    }

    argument_definitions = {
        'submit': {
            ('--custom-edition', '--ce'): {
                'help': 'Non-standard edition, e.g. "Final Cut"',
                'default': '',
            },
            ('--draft', '--dr'): {
                'help': 'Upload as draft',
                'action': 'store_true',
                # The default value must be None so CommandBase.get_options()
                # doesn't always overwrite the value with the config file value.
                'default': None,
            },
            ('--personal-rip', '--pr'): {
                'help': 'Tag as your own encode',
                'action': 'store_true',
            },
            ('--screenshots', '--ss'): {
                'help': ('How many screenshots to make '
                         f'(min={defaults["screenshots"].min}, '
                         f'max={defaults["screenshots"].max})'),
                'type': argtypes.number_of_screenshots(
                    min=defaults['screenshots'].min,
                    max=defaults['screenshots'].max,
                ),
            },
            ('--special', '--sp'): {
                'help': 'Tag as special episode, e.g. Christmas special (ignored for movie uploads)',
                'action': 'store_true',
            },
            ('--only-description', '--od'): {
                'help': 'Only generate description (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
            ('--only-title', '--ot'): {
                'help': 'Only generate title (do not upload anything)',
                'action': 'store_true',
                'group': 'generate-metadata',
            },
        },
    }



"""
Concrete :class:`~.base.TrackerJobsBase` subclass for BHD
"""

import functools
import io

from ... import __homepage__, __project_name__, errors, jobs, utils
from ..base import TrackerJobsBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class BhdTrackerJobs(TrackerJobsBase):

    @functools.cached_property
    def jobs_before_upload(self):
        return (
            # Background jobs
            self.create_torrent_job,
            self.mediainfo_job,
            self.screenshots_job,
            self.upload_screenshots_job,

            # Interactive jobs
            self.tmdb_job,
            self.imdb_job,
            self.release_name_job,
            self.category_job,
            self.type_job,
            self.source_job,
            self.description_job,
            self.scene_check_job,
            self.tags_job,
        )

    @property
    def isolated_jobs(self):
        """
        Sequence of job attribute names (e.g. "imdb_job") that were singled
        out by the user, e.g. with a CLI argument
        """
        if self.options.get('only_description', False):
            return self.get_job_and_dependencies(
                self.description_job,
                # `screenshots_job` is needed by `upload_screenshots_job`, but
                # `upload_screenshots_job` is a `QueueJobBase`, which doesn't
                # know anything about the job it gets input from and therefore
                # can't tells us that it needs `screenshots_job`.
                self.screenshots_job,
            )
        elif self.options.get('only_title', False):
            return self.get_job_and_dependencies(
                self.release_name_job,
                # `release_name_job` doesn't depend on `imdb_job` (or any other
                # webdb), but we want the correct name, year, etc in the release
                # name.
                self.imdb_job,
            )
        else:
            # Activate all jobs in jobs_before/after_upload
            return ()

    @functools.cached_property
    def category_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('category'),
            label='Category',
            precondition=self.make_precondition('category_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect=self.autodetect_category,
            choices=(
                ('Movie', '1'),
                ('TV', '2'),
            ),
            **self.common_job_args(),
        )

    _autodetect_category_map = {
        'Movie': lambda release_name: release_name.type is utils.release.ReleaseType.movie,
        'TV': lambda release_name: release_name.type in (utils.release.ReleaseType.season,
                                                         utils.release.ReleaseType.episode)
    }

    def autodetect_category(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting category: Approved release type: %r', approved_release_name.type)
        for label, is_match in self._autodetect_category_map.items():
            if is_match(approved_release_name):
                return label

    @functools.cached_property
    def type_job(self):
        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('type'),
            label='Type',
            precondition=self.make_precondition('type_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect_and_finish=self.autodetect_type,
            choices=(
                ('UHD 100', 'UHD 100'),
                ('UHD 66', 'UHD 66'),
                ('UHD 50', 'UHD 50'),
                ('UHD Remux', 'UHD Remux'),
                ('BD 50', 'BD 50'),
                ('BD 25', 'BD 25'),
                ('BD Remux', 'BD Remux'),
                ('2160p', '2160p'),
                ('1080p', '1080p'),
                ('1080i', '1080i'),
                ('720p', '720p'),
                ('576p', '576p'),
                ('540p', '540p'),
                ('DVD 9', 'DVD 9'),
                ('DVD 5', 'DVD 5'),
                ('DVD Remux', 'DVD Remux'),
                ('480p', '480p'),
                ('Other', 'Other'),
            ),
            focused='Other',
            **self.common_job_args(),
        )

    _autodetect_type_map = {
        'DVD 9': lambda release_name: release_name.source == 'DVD9',
        'DVD 5': lambda release_name: release_name.source == 'DVD5',
        'DVD Remux': lambda release_name: release_name.source == 'DVD Remux',
        '2160p': lambda release_name: release_name.resolution == '2160p',
        '1080p': lambda release_name: release_name.resolution == '1080p',
        '1080i': lambda release_name: release_name.resolution == '1080i',
        '720p': lambda release_name: release_name.resolution == '720p',
        '576p': lambda release_name: release_name.resolution == '576p',
        '540p': lambda release_name: release_name.resolution == '540p',
        '480p': lambda release_name: release_name.resolution == '480p',
    }

    def autodetect_type(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting type: Approved resolution: %r', approved_release_name.resolution)
        _log.debug('Autodetecting type: Approved source: %r', approved_release_name.source)
        for label, is_match in self._autodetect_type_map.items():
            if is_match(approved_release_name):
                return label

    @functools.cached_property
    def source_job(self):
        # 'output' signal is only emitted when job succeeds while 'finished'
        # signal is also emitted when job fails (e.g. when Ctrl-c is pressed)
        self.release_name_job.signal.register('output', self.autodetect_source)

        return jobs.dialog.ChoiceJob(
            name=self.get_job_name('source'),
            label='Source',
            precondition=self.make_precondition('source_job'),
            prejobs=(
                self.release_name_job,
            ),
            autodetect_and_finish=self.autodetect_source,
            choices=(
                ('Blu-ray', 'Blu-ray'),
                ('HD-DVD', 'HD-DVD'),
                ('WEB', 'WEB'),
                ('HDTV', 'HDTV'),
                ('DVD', 'DVD'),
            ),
            **self.common_job_args(),
        )

    # Map type_job labels to matchers
    _autodetect_source_map = {
        'Blu-ray': lambda release_name: 'BluRay' in release_name.source,
        'HD-DVD': lambda release_name: 'HD-DVD' in release_name.source,
        'WEB': lambda release_name: 'WEB' in release_name.source,
        'HDTV': lambda release_name: 'HDTV' in release_name.source,
        'DVD': lambda release_name: 'DVD' in release_name.source,
    }

    def autodetect_source(self, _):
        approved_release_name = self.release_name
        _log.debug('Autodetecting source: Approved source: %r', approved_release_name.source)
        for label, is_match in self._autodetect_source_map.items():
            if is_match(approved_release_name):
                return label

    @functools.cached_property
    def description_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('description'),
            label='Description',
            precondition=self.make_precondition('description_job'),
            prejobs=(
                self.upload_screenshots_job,
                self.mediainfo_job,
            ),
            text=self.generate_description,
            hidden=True,
            finish_on_success=True,
            read_only=True,
            **self.common_job_args(ignore_cache=True),
        )

    image_host_config = {
        'common': {'thumb_width': 350},
    }

    def generate_description(self):
        screenshot_tags = self._generate_description_screenshots()
        mediainfo_tags = self._generate_description_mediainfos()
        promotion_tag = (
            '[align=right][size=1]'
            f'Shared with [url={__homepage__}]{__project_name__}[/url]'
            '[/size][/align]'
        )

        description_parts = [
            '[center]\n'
            + screenshot_tags
            + '\n[/center]'
        ]
        if mediainfo_tags:
            # Include mediainfos if there are multiple (e.g. for IFO/VOB or if
            # `mediainfo_from_all_videos` is True)
            description_parts.append(
                '\n'
                + '[center][h3]Mediainfo[/h3][/center]\n'
                + mediainfo_tags
            )
        description_parts.append(f'\n\n{promotion_tag}')

        return ''.join(description_parts)

    def _generate_description_screenshots(self):
        assert self.upload_screenshots_job.is_finished
        return self.make_screenshots_grid(
            screenshots=self.upload_screenshots_job.uploaded_images,
            columns=2,
            horizontal_spacer='   ',
            vertical_spacer='\n\n',
        )

    def _generate_description_mediainfos(self):
        mediainfo_tags = []
        for video_filepath, mediainfo in self._description_mediainfos.items():
            filetitle = self.get_relative_file_path(video_filepath)
            mediainfo_tags.append(
                f'[hide={filetitle}][code]{mediainfo}[/code][/hide]'
            )
        return '\n'.join(mediainfo_tags)

    @property
    def _description_mediainfos(self):
        # If there are multiple files in the release, mediainfo for each file is
        # put in the description. Otherwise, the only needed mediainfo is passed
        # separately from the description
        assert self.mediainfo_job.is_finished
        mediainfos_by_file = self.mediainfo_job.mediainfos_by_file
        if len(mediainfos_by_file) >= 2:
            return {
                filepath: mediainfo
                for filepath, mediainfo in mediainfos_by_file.items()
                if mediainfo
            }
        else:
            return {}

    mediainfo_from_all_videos = False
    screenshots_from_all_videos = False

    @functools.cached_property
    def tags_job(self):
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('tags'),
            label='Tags',
            precondition=self.make_precondition('tags_job'),
            prejobs=(
                self.release_name_job,
                self.scene_check_job,
            ),
            text=self.generate_tags,
            finish_on_success=True,
            read_only=True,
            **self.common_job_args(),
        )

    async def generate_tags(self):
        assert self.release_name_job.is_finished
        assert self.scene_check_job.is_finished

        # Any additional tags separated by comma(s). (Commentary, 2in1, Hybrid,
        # OpenMatte, 2D3D, WEBRip, WEBDL, 3D, 4kRemaster, DualAudio, EnglishDub,
        # Personal, Scene, DigitalExtras, Extras)
        tags = []
        if 'WEBRip' in self.release_name.source:
            tags.append('WEBRip')
        elif 'WEB-DL' in self.release_name.source:
            tags.append('WEBDL')
        if 'Hybrid' in self.release_name.source:
            tags.append('Hybrid')
        if self.release_name.has_commentary:
            tags.append('Commentary')
        if self.release_name.has_dual_audio:
            tags.append('DualAudio')
        if 'Open Matte' in self.release_name.edition:
            tags.append('OpenMatte')
        if '2in1' in self.release_name.edition:
            tags.append('2in1')
        if '4k Remastered' in self.release_name.edition:
            tags.append('4kRemaster')
        if self.get_job_attribute(self.scene_check_job, 'is_scene_release'):
            tags.append('Scene')
        if self.options['personal_rip']:
            tags.append('Personal')

        # TODO: 2D3D
        # TODO: 3D
        # TODO: EnglishDub
        # TODO: DigitalExtras
        # TODO: Extras

        return '\n'.join(tags)

    @property
    def post_data(self):
        return {
            'name': self.get_job_output(self.release_name_job, slice=0),
            'category_id': self.get_job_attribute(self.category_job, 'choice'),
            'type': self.get_job_attribute(self.type_job, 'choice'),
            'source': self.get_job_attribute(self.source_job, 'choice'),
            'imdb_id': self.get_job_output(self.imdb_job, slice=0),
            'tmdb_id': self.get_job_output(self.tmdb_job, slice=0).split('/')[1],
            'description': self.get_job_output(self.description_job, slice=0),
            'edition': self.post_data_edition,
            'custom_edition': self.options['custom_edition'],
            'tags': ','.join(self.get_job_output(self.tags_job, slice=0).split('\n')),
            'nfo': self.post_data_nfo,
            'pack': self.post_data_pack,
            'sd': self.post_data_sd,
            'special': self.post_data_special,
            'anon': '1' if self.options['anonymous'] else '0',
            'live': '0' if self.options['draft'] else '1',
        }

    @functools.cached_property
    def post_data_edition(self):
        # The edition of the uploaded release. (Collector, Director, Extended,
        # Limited, Special, Theatrical, Uncut or Unrated)
        edition = self.release_name.edition
        _log.debug('Approved edition: %r', edition)
        if "Collector's Edition" in edition:
            return 'Collector'
        elif "Director's Cut" in edition:
            return 'Director'
        elif 'Extended Cut' in edition:
            return 'Extended'
        elif 'Limited' in edition:
            return 'Limited'
        elif 'Special Edition' in edition:
            return 'Special'
        elif 'Theatrical Cut' in edition:
            return 'Theatrical'
        elif 'Uncut' in edition or 'Uncensored' in edition:
            return 'Uncut'
        elif 'Unrated' in edition:
            return 'Unrated'

    @property
    def post_data_pack(self):
        # The TV pack flag for when the torrent contains a complete season.
        # (0 = No TV pack or 1 = TV Pack). Default is 0
        if self.release_name.type is utils.release.ReleaseType.season:
            return '1'
        else:
            return '0'

    @property
    def post_data_sd(self):
        # The SD flag. (0 = Not Standard Definition, 1 = Standard Definition).
        # Default is 0
        try:
            height = int(self.release_name.resolution[:-1])
        except ValueError:
            return '0'
        else:
            return '1' if height < 720 else '0'

    max_nfo_size = 500_000

    @property
    def post_data_nfo(self):
        try:
            return utils.string.read_nfo(self.content_path)
        except errors.ContentError as e:
            self.error(e)

    @property
    def post_data_special(self):
        # The TV special flag for when the torrent contains a TV special. (0 =
        # Not a TV special, 1 = TV Special). Default is 0
        if self.release_name.type is utils.release.ReleaseType.episode:
            if self.options['special']:
                return '1'
        return '0'

    # TODO
    # @property
    # def post_data_region(self):
    #     # The region in which the disc was released. Only for discs! (AUS,
    #     # CAN, CEE, CHN, ESP, EUR, FRA, GBR, GER, HKG, ITA, JPN, KOR, NOR,
    #     # NLD, RUS, TWN or USA)

    @property
    def torrent_filepath(self):
        return self.get_job_output(self.create_torrent_job, slice=0)

    @property
    def mediainfo_filehandle(self):
        mediainfo = self.get_job_output(self.mediainfo_job, slice=0)
        return io.BytesIO(bytes(mediainfo, 'utf-8'))



"""
Classes and functions that are used by all :class:`~.base.WebDbApiBase`
subclasses
"""

import os
import re

from .. import country, release, signal
from ..types import ReleaseType

import logging  # isort:skip
_log = logging.getLogger(__name__)


class Query:
    """
    Search query for databases like IMDb

    :param str title: Name of the movie or TV series
    :param type: :class:`~.types.ReleaseType` enum or one of its value names
    :param year: Year of release
    :type year: str or int
    :param str id: Known ID for a specific DB

    :raise ValueError: if an invalid argument is passed
    """

    @staticmethod
    def _normalize_title(title):
        return ' '.join(title.casefold().strip().split())

    _kwarg_defaults = {
        'year': None,
        'type': ReleaseType.unknown,
        'id': None,
    }

    def __init__(self, title='', **kwargs):
        for k in kwargs:
            if k not in self._kwarg_defaults:
                raise TypeError(f'Unkown argument: {k!r}')
        self._signal = signal.Signal()
        self._signal.add('changed')
        self.title = title
        self.type = kwargs.get('type', self._kwarg_defaults['type'])
        self.year = kwargs.get('year', self._kwarg_defaults['year'])
        self.id = kwargs.get('id', self._kwarg_defaults['id'])

    @property
    def type(self):
        """:class:`~.types.ReleaseType` value"""
        return self._type

    @type.setter
    def type(self, type):
        before = getattr(self, '_type', None)

        self._type = ReleaseType(type)

        if self._type != before:
            self.signal.emit('changed', self)

    @property
    def title(self):
        """Name of the movie or TV series"""
        return self._title

    @title.setter
    def title(self, title):
        before = getattr(self, '_title_normalized', None)

        self._title = str(title)
        self._title_normalized = self._normalize_title(self.title)

        if self._title_normalized != before:
            self.signal.emit('changed', self)

    @property
    def title_normalized(self):
        """Same as :attr:`title` but in stripped lower case with deduplicated spaces"""
        return self._title_normalized

    @property
    def year(self):
        """Year of release"""
        return self._year

    @year.setter
    def year(self, year):
        before = getattr(self, '_year', None)

        if not year:
            self._year = None
        else:
            try:
                year_int = int(year)
            except (TypeError, ValueError):
                raise ValueError(f'Invalid year: {year}')
            else:
                if not 1800 < year_int < 2100:
                    raise ValueError(f'Invalid year: {year}')
                else:
                    self._year = str(year_int)

        if self._year != before:
            self.signal.emit('changed', self)

    @property
    def id(self):
        """Known ID for a specific DB"""
        return self._id

    @id.setter
    def id(self, id):
        before = getattr(self, '_id', None)

        self._id = None if id is None else str(id)

        if self._id != before:
            self.signal.emit('changed', self)

    def update(self, query, silent=False):
        """
        Copy property values from other query

        :param query: :class:`Query` instance to copy values from
        :param bool silent: Whether to prevent any :attr:`signal` emissions
        """
        before = str(self)

        with self.signal.suspend('changed'):
            for attr in ('title', 'type', 'year', 'id'):
                other_value = getattr(query, attr)
                setattr(self, attr, other_value)

        if not silent and str(self) != before:
            self.signal.emit('changed', self)

    def copy(self, **updates):
        """
        Return new :class:`Query` instance with updated attributes

        :param updates: Updated attributes
        """
        kwargs = {
            'title': self.title,
            'type': self.type,
            'year': self.year,
            'id': self.id,
        }
        kwargs.update(updates)
        return type(self)(**kwargs)

    _types = {
        ReleaseType.movie: ('movie', 'film'),
        ReleaseType.season: ('season', 'series', 'tv', 'show', 'tvshow'),
        ReleaseType.episode: ('episode',),
    }
    _known_types = tuple(v for types in _types.values() for v in types)
    _kw_regex = {
        'year': r'year:(\d{4})',
        'type': rf'type:({"|".join(_known_types)})',
        'id': r'id:(\S+)',
    }

    @classmethod
    def from_string(cls, query):
        """
        Create instance from string

        The returned :class:`Query` is case-insensitive and has any superfluous
        whitespace removed.

        Keyword arguments are extracted by looking for ``"year:YEAR"``,
        ``"type:TYPE"`` and ``"id:ID"`` in `query` where ``YEAR`` is a
        four-digit number, ``TYPE`` is something like "movie", "film", "tv", etc
        and ``ID`` is a known ID for the DB this query is meant for.
        """
        def get_kwarg(string):
            for kw, regex in cls._kw_regex.items():
                match = re.search(f'^{regex}$', part)
                if match:
                    value = match.group(1)
                    if kw == 'type' and value in cls._types[ReleaseType.movie]:
                        return 'type', ReleaseType.movie
                    elif kw == 'type' and value in cls._types[ReleaseType.season]:
                        return 'type', ReleaseType.season
                    elif kw == 'type' and value in cls._types[ReleaseType.episode]:
                        return 'type', ReleaseType.episode
                    elif kw == 'year':
                        return 'year', value
                    elif kw == 'id':
                        return 'id', value
            return None, None

        # Extract "key:value" pairs (e.g. "year:2015")
        title = []
        kwargs = {}
        for part in str(query).strip().split():
            kw, value = get_kwarg(part)
            if (kw, value) != (None, None):
                kwargs[kw] = value
            else:
                title.append(part)
        if title:
            kwargs['title'] = ' '.join(title)
        return cls(**kwargs)

    @classmethod
    def from_release(cls, info):
        """
        Create instance from :class:`~.release.ReleaseInfo` or
        :class:`~.release.ReleaseName` instance
        """
        kwargs = {'title': info['title']}
        if info.get('year') and info.get('year') != 'UNKNOWN_YEAR':
            kwargs['year'] = info['year']
        if info.get('type'):
            kwargs['type'] = info['type']
        return cls(**kwargs)

    @classmethod
    def from_path(cls, path):
        """
        Create instance from file or directory name

        `path` is passed to :class:`~.release.ReleaseInfo` to get the
        arguments for instantiation.
        """
        info = release.ReleaseInfo(str(path))
        return cls.from_release(info)

    @classmethod
    def from_any(cls, obj):
        """
        Try to guess correct `from_…` method

        :raise TypeError: if `obj` is not supported
        """
        if isinstance(obj, cls):
            return obj
        elif isinstance(obj, str):
            if os.sep in obj or os.path.exists(obj):
                return cls.from_path(obj)
            else:
                return cls.from_string(obj)
        elif isinstance(obj, (release.ReleaseInfo, release.ReleaseName)):
            return cls.from_release(obj)
        else:
            raise TypeError(f'Unsupported type: {type(obj).__name__}: {obj!r}')

    @property
    def signal(self):
        """
        :class:`~.signal.Signal` instance

        Available signals:

        ``changed``
            Emitted after query parameters changed. Registered callbacks get the
            instance as a positional argument.
        """
        return self._signal

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                self.title_normalized == other.title_normalized
                and self.year == other.year
                and self.type is other.type
                and self.id == other.id
            )
        else:
            return NotImplemented

    def __str__(self):
        if self.id:
            return f'id:{self.id}'
        else:
            parts = [self.title]
            for attr in ('type', 'year', 'id'):
                value = getattr(self, attr)
                if value:
                    parts.append(f'{attr}:{value}')
            return ' '.join(parts)

    def __repr__(self):
        kwargs = ', '.join(
            f'{k}={v!r}'
            for k, v in (('title', self.title),
                         ('year', self.year),
                         ('type', self.type),
                         ('id', self.id))
            if v
        )
        return f'{type(self).__name__}({kwargs})'


class SearchResult:
    """
    Information about a search result

    Keyword arguments are available as attributes.

    Normal attributes:

    :param str id: ID for the relevant DB
    :param str title: Title of the movie or series
    :param str type: :class:`~.types.ReleaseType` value
    :param str url: Web page of the search result
    :param str year: Release year; for series this should be the year of the
        first airing of the first episode of the first season

    These attributes are coroutine functions that return the value when called
    with no arguments:

    :param cast: Short list of actor names
    :type cast: sequence of :class:`str`
    :param str countries: List of country names of origin
    :param genres: Short sequence of genres, e.g. `["horror", "comedy"]`
    :type genres: sequence of :class:`str`
    :param str directors: Sequence of directors
    :type directors: sequence of :class:`str`
    :param summary: Short text that describes the movie or series
    :param str title_english: English title of the movie or series
    :param str title_original: Original title of the movie or series

    The values of coroutine functions can be supplied via a coroutine function
    or as a plain object (:class:`str`, :class:`list`, etc).
    """

    def __init__(self, *, id, type, url, year, cast=(), countries=(), directors='',
                 genres=(), summary='', title, title_english='', title_original=''):
        self._info = {
            # Normal attributes functions
            'id': id,
            'title': str(title),
            'type': ReleaseType(type),
            'url': str(url),
            'year': str(year),
            # Coroutine functions
            'cast': self._ensure_async_getter('cast', cast),
            'countries': self._ensure_async_getter('countries', countries),
            'directors': self._ensure_async_getter('directors', directors),
            'genres': self._ensure_async_getter('genres', genres),
            'summary': self._ensure_async_getter('summary', summary),
            'title_english': self._ensure_async_getter('title_english', title_english),
            'title_original': self._ensure_async_getter('title_original', title_original),
        }

    def __getattr__(self, name):
        try:
            return self._info[name]
        except KeyError:
            raise AttributeError(name)

    def _ensure_async_getter(self, name, value):
        async def async_getter():
            # TODO: If Python 3.9 is no longer supported, use
            #       inspect.iscoroutinefunction() instead of callable().
            if callable(value):
                return self._upgrade_value(name, await value())
            else:
                return self._upgrade_value(name, value)

        async_getter.__qualname__ = f'async_get_{name}'
        return async_getter

    def _upgrade_value(self, name, value):
        if name == 'countries':
            return country.name(value)
        else:
            return value

    def __repr__(self):
        kwargs = ', '.join(f'{k}={v!r}' for k, v in self._info.items())
        return f'{type(self).__name__}({kwargs})'


class Person(str):
    """:class:`str` subclass with an `url` attribute"""

    def __new__(cls, name, url=''):
        obj = super().__new__(cls, name)
        obj.url = url
        return obj

    def __repr__(self):
        args = repr(str(self))
        if self.url:
            args += f', url={self.url!r}'
        return f'{type(self).__name__}({args})'



"""
Concrete :class:`~.base.TrackerBase` subclass for PTP
"""

import math
import re
import urllib

from ... import errors, utils
from ..base import TrackerBase
from .config import PtpTrackerConfig
from .jobs import PtpTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class PtpTracker(TrackerBase):
    name = 'ptp'
    label = 'PTP'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.current_section}. Login Credentials\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.ptp.username USERNAME\n'
        '   {howto.current_section}.2 $ upsies set trackers.ptp.password PASSWORD\n'
        '{howto.bump_section}'
        '\n'
        '{howto.current_section}. Announce URL\n'
        '\n'
        '   The announce URL is required because the passkey is used for 2FA.\n'
        '\n'
        '   {howto.current_section}.1 $ upsies set trackers.ptp.announce_url ANNOUNCE_URL\n'
        '{howto.bump_section}'
        '\n'
        '{howto.current_section}. Screenshots\n'
        '\n'
        '   {howto.current_section}.1 Configure ptpimg.me API\n'
        '       $ upsies upload-images ptpimg --help\n'
        '\n'
        '   {howto.current_section}.2 Specify how many screenshots to make. (optional)\n'
        '       $ upsies set trackers.ptp.screenshots_from_movie NUMBER_OF_MOVIE_SCREENSHOTS\n'
        '       $ upsies set trackers.ptp.screenshots_from_episode NUMBER_OF_EPISODE_SCREENSHOTS\n'
        '{howto.bump_section}'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerConfig = PtpTrackerConfig
    TrackerJobs = PtpTrackerJobs

    @property
    def _base_url(self):
        return self.options['base_url']

    @property
    def _ajax_url(self):
        return urllib.parse.urljoin(self._base_url, '/ajax.php')

    @property
    def _logout_url(self):
        return urllib.parse.urljoin(self._base_url, '/logout.php')

    @property
    def _upload_url(self):
        return urllib.parse.urljoin(self._base_url, '/upload.php')

    @property
    def _torrents_url(self):
        return urllib.parse.urljoin(self._base_url, '/torrents.php')

    @property
    def _announce_url(self):
        announce_url = self.options['announce_url']
        if not announce_url:
            raise errors.AnnounceUrlNotSetError(tracker=self)
        else:
            return self.options['announce_url']

    @property
    def _passkey(self):
        # Needed for logging in with ajax.php
        match = re.search(r'.*/([a-zA-Z0-9]+)/announce', self._announce_url)
        if match:
            return match.group(1)
        else:
            raise RuntimeError(f'Failed to find passkey in announce URL: {self._announce_url}')

    async def _request(self, method, *args, error_prefix='', **kwargs):
        # Because HTTP errors (e.g. 404) are raised, we treat RequestErrrors as
        # normal response so we can get the message from the HTML.
        try:
            # `method` is "GET" or "POST"
            response = await getattr(utils.http, method.lower())(
                *args,
                user_agent=True,
                follow_redirects=False,
                **kwargs,
            )
        except errors.RequestError as e:
            response = e

        # Get error from regular exception (e.g. "Connection refused") or the
        # HTML in response.
        try:
            self._maybe_raise_error(response)
        except errors.RequestError as e:
            # Prepend error_prefix to explain the general nature of the error.
            if error_prefix:
                raise errors.RequestError(f'{error_prefix}: {e}')
            else:
                raise e
        else:
            return response

    def _maybe_raise_error(self, response_or_request_error):
        # utils.http.get()/post() raise RequestError on HTTP status codes, but
        # we want to get the error message from the response text.
        # _maybe_raise_error_from_*() handle Response and RequestError.
        self._maybe_raise_error_from_json(response_or_request_error)
        self._maybe_raise_error_from_html(response_or_request_error)

        # If we got a RequestError and we didn't find an error message in the
        # text, we raise it. This handles any real RequestErrors, like
        # "Connection refused". We also raise any other exception so _request()
        # doesn't return it as a regular response.
        if isinstance(response_or_request_error, BaseException):
            raise response_or_request_error

    def _maybe_raise_error_from_json(self, response_or_request_error):
        # Get error message from ajax.php JSON Response or RequestError
        try:
            json = response_or_request_error.json()
        except errors.RequestError:
            # Response or RequestError is not JSON
            pass
        else:
            if (
                isinstance(json, dict)
                and json.get('Result') == 'Error'
                and json.get('Message')
            ):
                raise errors.RequestError(utils.html.as_text(json['Message']))

    def _maybe_raise_error_from_html(self, response_or_request_error):
        # Only attempt to find an error message if this looks like HTML. This
        # prevents a warning from bs4 about parsing non-HTML.
        text = str(response_or_request_error)
        if all(c in text for c in '<>\n'):
            doc = utils.html.parse(text)
            try:
                error_header_tag = doc.select('#content .page__title', string=re.compile(r'(?i:error)'))
                error_container_tag = error_header_tag[0].parent
                error_msg_tag = error_container_tag.find('div', attrs={'class': 'panel__body'})
                error_msg = error_msg_tag.get_text().strip()
                if error_msg:
                    raise errors.RequestError(error_msg)
            except (AttributeError, IndexError):
                # No error message found
                pass

    def _get_anti_csrf_token(self, response):
        json = response.json()
        try:
            return json['AntiCsrfToken']
        except KeyError:
            _log.debug('Failed to find "AntiCsrfToken": %r', response)
            raise

    async def _get_auth(self):
        assert self.is_logged_in
        response = await self._request('GET', self._base_url)
        doc = utils.html.parse(response)
        auth_regex = re.compile(r'logout\.php\?.*\bauth=([0-9a-fA-F]+)')
        logout_link_tag = doc.find('a', href=auth_regex)
        if logout_link_tag:
            logout_link_href = logout_link_tag['href']
            match = auth_regex.search(logout_link_href)
            return match.group(1)

        raise RuntimeError('Could not find auth')

    async def _login(self):
        if not self.options.get('username'):
            raise errors.RequestError('Login failed: No username configured')
        elif not self.options.get('password'):
            raise errors.RequestError('Login failed: No password configured')

        _log.debug('%s: Logging in as %r', self.name, self.options['username'])
        post_data = {
            'username': self.options['username'],
            'password': self.options['password'],
            'passkey': self._passkey,
            # 'keeplogged': '1',
        }
        response = await self._request(
            method='POST',
            url=f'{self._ajax_url}?action=login',
            data=post_data,
            error_prefix='Login failed',

            # TODO: Remove this before release
            debug_file='http_dumps.ptp/ptp_login',
        )
        self._anti_csrf_token = self._get_anti_csrf_token(response)

    async def _logout(self):
        try:
            _log.debug('%s: Logging out', self.name)
            await self._request(
                method='GET',
                url=self._logout_url,
                params={'auth': await self._get_auth()},
                error_prefix='Logout failed',

                # TODO: Remove this before release
                debug_file='http_dumps.ptp/ptp_logout',
            )

        finally:
            if hasattr(self, '_anti_csrf_token'):
                delattr(self, '_anti_csrf_token')

    async def get_announce_url(self):
        return self._announce_url

    async def upload(self, tracker_jobs):
        post_data = tracker_jobs.post_data.copy()
        post_data['AntiCsrfToken'] = self._anti_csrf_token

        _log.debug('POSTing data:')
        for k, v in post_data.items():
            _log.debug(' * %s = %s', k, v)

        post_files = {
            'file_input': {
                'file': tracker_jobs.torrent_filepath,
                'mimetype': 'application/x-bittorrent',
            },
        }
        _log.debug('POSTing files: %r', post_files)

        response = await utils.http.post(
            url=self._upload_url,
            cache=False,
            user_agent=True,
            data=post_data,
            files=post_files,
            # Ignore the HTTP redirect (should be 302 Found) so we can get the
            # torrent URL from the "Location" response header
            follow_redirects=False,

            # TODO: Remove this before release
            debug_file='http_dumps.ptp/ptp_upload',
        )

        # # Read previously received response for debugging
        # response_filepath = 'ptp_upload.response.content'
        # response = utils.http.Response(
        #     text=open(response_filepath).read(),
        #     bytes=open(response_filepath).read().encode('utf8'),
        # )

        return self._handle_upload_response(response)

    def _handle_upload_response(self, response):
        # "Location" header should contain the uploaded torrent's URL
        _log.debug('Upload response headers: %r', response.headers)
        location = response.headers.get('Location')
        _log.debug('Upload response location: %r', location)
        if location:
            torrent_page_url = urllib.parse.urljoin(self.options['base_url'], location)
            # Redirect URL should start with "https://.../torrents.php"
            if torrent_page_url.startswith(self._torrents_url):
                return torrent_page_url

        # Find error message in HTML
        doc = utils.html.parse(response)

        # Find and raise error message
        alert_tag = doc.find(class_='alert')
        if alert_tag:
            msg = utils.html.as_text(alert_tag)
            raise errors.RequestError(f'Upload failed: {msg}')

        # Failed to find error message
        dump_filepath = 'ptp_upload_failed.html'
        utils.html.dump(response, dump_filepath)
        raise errors.RequestError(f'Failed to interpret response (see {dump_filepath})')

    def normalize_imdb_id(self, imdb_id):
        """
        Format IMDb ID for PTP

        PTP expects 7-characters, right-padded with "0" and without the leading
        "tt".

        If `imdb_id` is falsy (e.g. `None`, empty string, etc), or if it isn't
        an IMDb ID, return "0".
        """
        imdb_id = str(imdb_id)
        match = re.search(r'^(?:tt|)(\d+)$', imdb_id)
        if match:
            imdb_id_digits = match.group(1)
            if set(imdb_id_digits) != {'0'}:
                return match.group(1).rjust(7, '0')
        return '0'

    async def get_ptp_group_id_by_imdb_id(self, imdb_id):
        """
        Convert IMDb ID to PTP group ID

        Any :class:`~.RequestError` is caught and passed to
        :meth:`.TrackerBase.error`.

        :return: PTP group ID or `None` if PTP doesn't have a group for
            `imdb_id`
        """
        _log.debug('%s: Fetching PTP group ID', imdb_id)
        try:
            await self.login()
            response = await self._request(
                method='GET',
                url=self._torrents_url,
                params={
                    'imdb': self.normalize_imdb_id(imdb_id),
                    'json': '1',
                },
                cache=True,
            )

        except errors.RequestError as e:
            self.error(e)

        else:
            match = re.search(r'id=(\d+)', response.headers.get('location', ''))
            if match:
                _log.debug('%s: PTP group ID: %s', imdb_id, match.group(1))
                return match.group(1)
            else:
                _log.debug('%s: No PTP group ID', imdb_id)

    async def get_ptp_metadata(self, imdb_id, key=None):
        """
        Get metadata from PTP website

        Any :class:`~.RequestError` is caught and passed to
        :meth:`.TrackerBase.error`.

        :param imdb_id: IMDb ID (e.g. ``tt123456``)
        """
        if key:
            _log.debug('%s: Fetching %s from PTP', imdb_id, key)
        else:
            _log.debug('%s: Fetching metadata from PTP', imdb_id)

        assert imdb_id, imdb_id

        try:
            await self.login()
            response = await self._request(
                method='GET',
                url=self._ajax_url,
                params={
                    'action': 'torrent_info',
                    'imdb': self.normalize_imdb_id(imdb_id),
                },
                cache=True,
            )
            # Raise RequestError if response is not valid JSON
            results = response.json()

        except errors.RequestError as e:
            self.error(e)

        else:
            assert len(results) == 1
            if key is None:
                _log.debug('%s: PTP metadata: %s', imdb_id, results[0])
                return results[0]
            else:
                _log.debug('%s: PTP %s: %s', imdb_id, key, results[0][key])
                return results[0][key]

    @staticmethod
    def calculate_piece_size(bytes):
        """
        Return the recommended piece size for a given content size

        :param bytes: Torrent's content size

            .. note:: Remember to exclude files that exist but won't be in the
                torrent (e.g. extras or ``*.nfo``).
        """
        exponent = math.ceil(math.log2(bytes / 1050))
        # Allowed piece size range: 32 KiB ... 16 MiB
        exponent = min(24, max(15, exponent))
        return int(math.pow(2, exponent))

    @staticmethod
    def calculate_piece_size_min_max(bytes):
        """
        Return the allowed minimum and maximum piece size for a given
        content size

        :param bytes: Torrent's content size

            .. note:: Remember to exclude files that exist but won't be in the
                torrent (e.g. extras or ``*.nfo``).

        :raise ValueError: if `bytes` is negative or otherwise unexpected
        """
        # All numbers are from PTP's piece size chart.
        # (MIN_SIZE, MAX_SIZE): (PIECE_SIZE_MIN, PIECE_SIZE_MAX)
        allowed_piece_sizes = {
            (0, 147 * 1024 * 1024 - 1): (2**15, 2**16),
            (147 * 1024 * 1024, 256 * 1024 * 1024 - 1): (2**16, 2**19),
            (256 * 1024 * 1024, 533 * 1024 * 1024 - 1): (2**17, 2**20),
            (533 * 1024 * 1024, 1.08 * 1024 * 1024 * 1024 - 1): (2**18, 2**21),
            (1.08 * 1024 * 1024 * 1024, 2.24 * 1024 * 1024 * 1024 - 1): (2**19, 2**22),
            (2.24 * 1024 * 1024 * 1024, 4.65 * 1024 * 1024 * 1024 - 1): (2**20, 2**23),
            (4.65 * 1024 * 1024 * 1024, 8.04 * 1024 * 1024 * 1024 - 1): (2**21, 2**24),
            (8.04 * 1024 * 1024 * 1024, 148.75 * 1024 * 1024 * 1024 - 1): (2**22, 2**24),
            (148.75 * 1024 * 1024 * 1024, 257.04 * 1024 * 1024 * 1024 - 1): (2**23, 2**24),
            (257.04 * 1024 * 1024 * 1024, float('inf')): (2**24, 2**24),
        }

        for (min_size, max_size), (piece_size_min, piece_size_max) in allowed_piece_sizes.items():
            if min_size <= bytes <= max_size:
                return piece_size_min, piece_size_max

        raise ValueError(f'Unexpected size: {bytes!r}')



import re
from unittest.mock import call

import pytest

from upsies.utils import release


def test_Episodes_init(mocker):
    set_mock = mocker.patch('upsies.utils.release.Episodes._set')
    release.Episodes('foo', bar='baz')
    assert set_mock.call_args_list == [call('foo', bar='baz')]


@pytest.mark.parametrize(
    argnames='clear',
    argvalues=(None, False, True),
)
def test_Episodes_update(clear, mocker):
    set_mock = mocker.patch('upsies.utils.release.Episodes._set')
    e = release.Episodes('initial', arg='uments')
    initial_e = dict(e)

    if clear is not None:
        e.update('foo', bar='baz', clear=clear)
    else:
        e.update('foo', bar='baz')

    assert set_mock.call_args_list == [
        call('initial', arg='uments'),
        call('foo', bar='baz'),
    ]

    if clear:
        assert e == {}
    else:
        assert e == initial_e


@pytest.mark.parametrize(
    argnames='initial, update_args, update_kwargs, exp_result',
    argvalues=(
        # Setting positional arguments
        (
            # Initial episodes
            {'1': ['2', '3'], '4': [], '': [6, '6', '5']},
            # Positional arguments
            ({'1': ['20', '30'], '40': [], '': ['60', 5, '50'], '10': '100'},),
            # Keyword arguments
            {},
            # Combined (initial + positional + keyword)
            {'1': ['2', '3', '20', '30'], '4': [], '40': [], '': ['5', '6', '50', '60'], '10': ['100']},
        ),
        # Setting keyword arguments
        (
            # Initial episodes
            {'1': ['2', '3'], '4': [], '': [6, '6', '5']},
            # Positional arguments
            (),
            # Keyword arguments
            {'S01': ['E20', 'E30'], 'S40': [], 'S': ['E60', 5, 'E50'], 'S10': 'E100'},
            # Combined (initial + positional + keyword)
            {'1': ['2', '3', '20', '30'], '4': [], '40': [], '': ['5', '6', '50', '60'], '10': ['100']},
        ),
        # Setting positional and keyword arguments
        (
            # Initial episodes
            {'1': ['2', '3'], '4': [], '': [6, '6', '5']},
            # Positional arguments
            ({'1': ['20', '40'], '': ['60'], '10': ['100']},),
            # Keyword arguments
            {'S01': ['E30', '50'], 'S40': [], 'S': [5, 'E50']},
            # Combined (initial + positional + keyword)
            {'1': ['2', '3', '20', '30', '40', '50'], '4': [], '40': [], '': ['5', '6', '50', '60'], '10': ['100']},
        ),
    ),
    ids=lambda v: str(v),
)
def test_Episodes_set(initial, update_args, update_kwargs, exp_result, mocker):
    e = release.Episodes(initial)
    e._set(*update_args, **update_kwargs)
    assert e == exp_result


def test_Episodes_normalize_season(mocker):
    e = release.Episodes()
    mocker.patch.object(e, '_normalize', return_value='mock number')
    season = e._normalize_season('foo')
    assert season == 'mock number'
    assert e._normalize.call_args_list == [call('foo', name='season', prefix='S', empty_string_ok=True)]


def test_Episodes_normalize_episode(mocker):
    e = release.Episodes()
    mocker.patch.object(e, '_normalize', return_value='mock number')
    episode = e._normalize_episode('foo')
    assert episode == 'mock number'
    assert e._normalize.call_args_list == [call('foo', name='episode', prefix='E', empty_string_ok=False)]


@pytest.mark.parametrize(
    argnames='value, name, prefix, empty_string_ok, exp_return_value, exp_exception',
    argvalues=(
        # Integer
        (-1, 'potatoe', 'P', None, None, TypeError('Invalid potatoe: -1')),
        (0, 'potatoe', 'P', None, '0', None),
        (1, 'potatoe', 'P', None, '1', None),

        # String
        ('-1', 'potatoe', 'P', None, None, TypeError("Invalid potatoe: '-1'")),
        ('0', 'potatoe', 'P', None, '0', None),
        ('1', 'potatoe', 'P', None, '1', None),
        ('P2', 'potatoe', 'P', None, '2', None),
        ('p3', 'potatoe', 'P', None, '3', None),
        ('P04', 'potatoe', 'P', None, '4', None),
        ('p05', 'potatoe', 'P', None, '5', None),
        ('PP6', 'potatoe', 'P', None, None, TypeError("Invalid potatoe: 'P6'")),

        # Empty string
        ('', 'potatoe', 'P', None, None, TypeError("Invalid potatoe: ''")),
        ('', 'potatoe', 'P', False, None, TypeError("Invalid potatoe: ''")),
        ('', 'potatoe', 'P', True, '', None),
        ('P', 'potatoe', 'P', None, None, TypeError("Invalid potatoe: ''")),
        ('P', 'potatoe', 'P', False, None, TypeError("Invalid potatoe: ''")),
        ('P', 'potatoe', 'P', True, '', None),
        ('p', 'potatoe', 'P', None, None, TypeError("Invalid potatoe: ''")),
        ('p', 'potatoe', 'P', False, None, TypeError("Invalid potatoe: ''")),
        ('p', 'potatoe', 'P', True, '', None),
        ('pp', 'potatoe', 'P', True, None, TypeError("Invalid potatoe: 'p'")),
        ('PP', 'potatoe', 'P', True, None, TypeError("Invalid potatoe: 'P'")),

        # Other types
        (1.2, 'potatoe', 'P', None, None, TypeError("Invalid potatoe: 1.2")),
        ([1, 2, 3], 'potatoe', 'P', None, None, TypeError("Invalid potatoe: [1, 2, 3]")),
    ),
    ids=lambda v: str(v),
)
def test_Episodes_normalize(value, name, prefix, empty_string_ok, exp_return_value, exp_exception):
    e = release.Episodes()

    kwargs = {
        'name': name,
        'prefix': prefix,
        'empty_string_ok': empty_string_ok,
    }
    # Remove `None` values
    for k, v in tuple(kwargs.items()):
        if v is None:
            del kwargs[k]

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            print(e._normalize(value, **kwargs))
    else:
        return_value = e._normalize(value, **kwargs)
        assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames=('string', 'exp'),
    argvalues=(
        ('foo.E01 bar', {'': ['1',]}),
        ('foo.E00 bar', {'': ['0',]}),
        ('foo E01E2.bar', {'': ['1', '2']}),
        ('foo E00E1.bar', {'': ['0', '1']}),
        ('foo S02E03.bar', {'2': ['3']}),
        ('foo S02E00.bar', {'2': ['0']}),
        ('foo S00E03.bar', {'0': ['3']}),
        ('foo.bar.E01E2S03', {'': ['1', '2'], '3': []}),
        ('foo-E01E2S03E04E05.bar.baz', {'': ['1', '2'], '3': ['4', '5']}),
        ('foo.S09E08S03E06 bar S9E1 baz', {'9': ['1', '8',], '3': ['6',]}),
        ('123 E01S03E06.456.E02', {'': ['1'], '3': ['2', '6']}),
        ('hello S03E14-E15 goodbye', {'3': ['14', '15',]}),
        ('this-s02e01-that', {'2': ['1']}),
        ('this-s02e10-E20-that', {'2': ['10', '20']}),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.parametrize('change_string', (None, str.lower, str.upper))
def test_Episodes_from_string(string, change_string, exp):
    if change_string:
        string = change_string(string)
    assert release.Episodes.from_string(string) == exp


@pytest.mark.parametrize(
    argnames=('sequence', 'exp'),
    argvalues=(
        (('foo.S01E01.bar', 'foo.S01E02.baz'), {'1': ['1', '2']}),
        (('foo.S02E02E01.bar', 'foo.S01E01.bar', 'foo.S02E03.baz'), {'1': ['1',], '2': ['1', '2', '3']}),
    ),
    ids=lambda v: str(v),
)
def test_Episodes_from_sequence(sequence, exp):
    assert release.Episodes.from_sequence(sequence) == exp


@pytest.mark.parametrize(
    argnames=('mapping', 'string'),
    argvalues=(
        ({'': [1,]}, 'E01'),
        ({'': [1, 10]}, 'E01E10'),
        ({4: [1, 10]}, 'S04E01E10'),
        ({2: [4,], 1: [2, 3]}, 'S01E02E03S02E04'),
        ({1: [2, 3], 2: [4,], '': [5,]}, 'E05S01E02E03S02E04'),
    ),
    ids=lambda v: str(v),
)
def test_Episodes_str(mapping, string):
    assert str(release.Episodes(mapping)) == string


def test_Episodes_repr():
    x = release.Episodes({5: [3, 6], '': [1, 2], 6: []})
    assert repr(x) == "Episodes({'5': ['3', '6'], '': ['1', '2'], '6': []})"


@pytest.mark.parametrize(
    argnames=('release_name', 'exp_value'),
    argvalues=(
        ('Foo.S01.720p.BluRay.x264-ASDF', True),
        ('Foo.S01E02.720p.BluRay.x264-ASDF', True),
        ('Foo.S01S02.720p.BluRay.x264-ASDF', True),
        ('Foo.S01E01E02.720p.BluRay.x264-ASDF', True),
        ('Foo.2015.S01.720p.BluRay.x264-ASDF', True),
        ('Foo.2015.S01E02.720p.BluRay.x264-ASDF', True),
        ('Foo.2015.S01S02.720p.BluRay.x264-ASDF', True),
        ('Foo.2015.S01E01E02.720p.BluRay.x264-ASDF', True),
        ('Foo.2015.720p.BluRay.x264-ASDF', False),
    ),
    ids=lambda v: str(v),
)
def test_has_episodes_info(release_name, exp_value):
    assert release.Episodes.has_episodes_info(release_name) is exp_value
    assert release.Episodes.has_episodes_info(release_name.lower()) is exp_value


@pytest.mark.parametrize(
    argnames=('string', 'exp_value'),
    argvalues=(
        ('S01', True),
        ('S01E02', True),
        ('S01E02E03', True),
        ('S01S02', True),
        ('S01S02E03', True),
        ('S01E02S03E04', True),
        ('S01E02E03S04E05S06', True),
        ('E01E02', True),
        ('.S01', False),
        ('S01.', False),
    ),
    ids=lambda v: str(v),
)
def test_is_episodes_info(string, exp_value):
    assert release.Episodes.is_episodes_info(string) is exp_value
    assert release.Episodes.is_episodes_info(string.lower()) is exp_value


@pytest.mark.parametrize(
    argnames=('episodes', 'exp_episodes'),
    argvalues=(
        ({}, {}),
        ({'': ('1',)}, {}),
        ({'1': (), '2': ('3',)}, {'1': (), '2': ()}),
    ),
    ids=lambda v: str(v),
)
def test_remove_specific_episodes(episodes, exp_episodes):
    e = release.Episodes(episodes)
    e.remove_specific_episodes()
    assert e == exp_episodes



from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.jobs.submit import SubmitJob
from upsies.trackers import TrackerBase, TrackerJobsBase


@pytest.fixture
def job(tmp_path, tracker, tracker_jobs):
    return SubmitJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        tracker=tracker,
        tracker_jobs=tracker_jobs,
    )

@pytest.fixture
def tracker():
    class TestTracker(TrackerBase):
        name = 'test'
        label = 'TeST'
        TrackerConfig = 'tracker config class'
        TrackerJobs = 'tracker jobs class'
        login = AsyncMock()
        logout = AsyncMock()
        _login = AsyncMock()
        _logout = AsyncMock()
        is_logged_in = PropertyMock()
        get_announce_url = AsyncMock()
        upload = AsyncMock()

    return TestTracker()

@pytest.fixture
def tracker_jobs(tracker):
    kwargs = {
        'content_path': 'content/path',
        'tracker': tracker,
        'btclient': Mock(),
        'torrent_destination': 'torrent/destination',
        'image_host': 'image host',
        'common_job_args': {},
    }

    class TestTrackerJobs(TrackerJobsBase):
        def __init__(self, **kw):
            return super().__init__(**{**kwargs, **kw})

        jobs_before_upload = PropertyMock()
        jobs_after_upload = PropertyMock()
        submission_ok = PropertyMock(return_value=True)

    return TestTrackerJobs()


def test_cache_id(job):
    assert job.cache_id == job._tracker.name


@pytest.mark.asyncio
async def test_initialize_creates_jobs_after_upload(tmp_path, tracker, tracker_jobs, mocker):
    logged = []

    def log(msg, return_value):
        logged.append(msg)
        return return_value

    type(tracker_jobs).jobs_after_upload = PropertyMock(
        side_effect=lambda: log('Creating jobs_after_upload', [Mock(), Mock()]),
    )
    SubmitJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        tracker=tracker,
        tracker_jobs=tracker_jobs,
    )
    assert logged == ['Creating jobs_after_upload']


@pytest.mark.asyncio
async def test_initialize_registers_to_tracker_signals(tmp_path, tracker, tracker_jobs):
    job = SubmitJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        tracker=tracker,
        tracker_jobs=tracker_jobs,
    )
    for t in (tracker, tracker_jobs):
        assert job.warn in t.signal.signals['warning']
        assert job.error in t.signal.signals['error']
        assert job.exception in t.signal.signals['exception']

@pytest.mark.asyncio
async def test_initialize_starts_jobs_after_upload_on_output(tmp_path, tracker, tracker_jobs, mocker):
    type(tracker_jobs).jobs_after_upload = PropertyMock(
        return_value=[Mock(), Mock(), Mock()],
    )
    job = SubmitJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        tracker=tracker,
        tracker_jobs=tracker_jobs,
    )
    assert len(job.jobs_after_upload) > 0
    for j in job.jobs_after_upload:
        assert j.start.call_args_list == []
    job.send('foo')
    for j in job.jobs_after_upload:
        assert j.start.call_args_list == [call()]


def test_execute_adds_run_jobs_task(job, mocker):
    mocker.patch.object(job, 'attach_task')
    mocker.patch.object(job, '_run_jobs', Mock())
    assert job.attach_task.call_args_list == []
    job.execute()
    assert job.attach_task.call_args_list == [call(job._run_jobs.return_value)]
    assert job._run_jobs.call_args_list == [call()]

def test_execute_does_nothing_if_job_is_finished_task(job, mocker):
    mocker.patch.object(job, 'attach_task')
    mocker.patch.object(job, '_run_jobs')
    assert job.attach_task.call_args_list == []
    job.finish()
    job.execute()
    assert job.attach_task.call_args_list == []


@pytest.mark.asyncio
async def test_run_jobs_waits_for_jobs_until_all_are_finished(job, mocker):
    class MockJob:
        def __init__(self, name, is_finished_values):
            self.name = name
            self.exit_code = 0
            self._is_finished_values = list(is_finished_values)
            self._is_finished = self._is_finished_values.pop(0)
            self.wait_calls = 0

        async def wait(self):
            if self._is_finished_values:
                self.wait_calls += 1
                self._is_finished = self._is_finished_values.pop(0)

        @property
        def is_finished(self):
            return self._is_finished

    jobs = {
        'a': MockJob('a', (False, False, True)),
        'b': MockJob('b', (False, True)),
        'c': MockJob('c', (False, False, False, True)),
    }
    mocker.patch.object(type(job), 'jobs_before_upload', PropertyMock(side_effect=(
        (jobs['a'],),
        (jobs['a'], jobs['b']),
        (jobs['a'], jobs['b'], jobs['c']),
        (jobs['a'], jobs['b'], jobs['c']),
        (jobs['a'], jobs['b'], jobs['c']),
        (jobs['a'], jobs['b'], jobs['c']),
    )))
    await job._run_jobs()
    assert jobs['a'].wait_calls == 2
    assert jobs['b'].wait_calls == 1
    assert jobs['c'].wait_calls == 3

@pytest.mark.asyncio
async def test_run_jobs_waits_for_tracker_tasks(job, mocker):
    mocks = AsyncMock()
    mocks.attach_mock(mocker.patch.object(job._tracker, 'await_tasks', AsyncMock()), 'await_tasks')

    class MockJob:
        def __init__(self, name, is_finished_values):
            self.name = name
            self._is_finished_values = list(is_finished_values)
            self._is_finished = self._is_finished_values.pop(0)
            self.wait = getattr(mocks, f'wait_{name}')
            self.wait.side_effect = self._wait

        async def _wait(self):
            if self._is_finished_values:
                self._is_finished = self._is_finished_values.pop(0)

        @property
        def is_finished(self):
            return self._is_finished

    mocker.patch.object(type(job), 'jobs_before_upload', PropertyMock(return_value=(
        MockJob('a', (False, True)),
        MockJob('b', (False, True)),
        MockJob('c', (False, True)),
    )))

    await job._run_jobs()

    assert mocks.mock_calls == [
        call.wait_a(),
        call.wait_b(),
        call.wait_c(),
        call.await_tasks(),
    ]

@pytest.mark.parametrize('submission_ok, exp_calls', ((True, [call()]), (False, [])))
@pytest.mark.asyncio
async def test_run_jobs_submits_if_submission_ok(submission_ok, exp_calls, job, mocker):
    mocker.patch.object(job, '_submit', AsyncMock())
    mocker.patch.object(type(job._tracker_jobs), 'submission_ok', PropertyMock(return_value=submission_ok))
    assert not job.is_finished
    await job._run_jobs()
    assert job._submit.call_args_list == exp_calls
    assert job.is_finished


@pytest.mark.parametrize('method', ('login', 'logout', 'upload'))
@pytest.mark.asyncio
async def test_submit_handles_RequestError_from_tracker_coro(method, job):
    setattr(
        job._tracker,
        method,
        AsyncMock(side_effect=errors.RequestError(f'{method}: No connection')),
    )
    assert job.output == ()
    assert job.errors == ()
    await job._submit()
    if method == 'logout':
        assert job.output == (str(job._tracker.upload.return_value),)
    else:
        assert job.output == ()
    assert job.errors == (errors.RequestError(f'{method}: No connection'),)
    if method in ('upload', 'logout'):
        assert job._tracker.logout.call_args_list == [call()]
    else:
        assert job._tracker.logout.call_args_list == []

@pytest.mark.asyncio
async def test_submit_calls_methods_and_callbacks_in_correct_order(job, mocker):
    mocks = Mock()
    mocks.attach_mock(job._tracker.login, 'login')
    mocks.attach_mock(job._tracker.logout, 'logout')
    mocks.attach_mock(job._tracker.upload, 'upload')
    job.signal.register('logging_in', mocks.logging_in_cb)
    job.signal.register('logged_in', mocks.logged_in_cb)
    job.signal.register('uploading', mocks.uploading_cb)
    job.signal.register('uploaded', mocks.uploaded_cb)
    job.signal.register('logging_out', mocks.logging_out_cb)
    job.signal.register('logged_out', mocks.logged_out_cb)
    await job._submit()
    assert mocks.method_calls == [
        call.logging_in_cb(),
        call.login(),
        call.logged_in_cb(),

        call.uploading_cb(),
        call.upload(job._tracker_jobs),
        call.uploaded_cb(),

        call.logging_out_cb(),
        call.logout(),
        call.logged_out_cb(),
    ]

@pytest.mark.parametrize(
    argnames='torrent_url, exp_output',
    argvalues=(
        ('http://torrent.url', ('http://torrent.url',)),
        ('', ()),
        (None, ()),
    ),
)
@pytest.mark.asyncio
async def test_submit_sends_upload_return_value_as_output(torrent_url, exp_output, job):
    job._tracker.upload.return_value = torrent_url
    await job._submit()
    assert job.output == exp_output

@pytest.mark.asyncio
async def test_submit_sets_info_property_to_current_status(job):
    infos = [
        'Logging in',
        'Logged in',
        'Uploading',
        'Uploaded',
        'Logging out',
        '',
    ]

    def info_cb():
        assert job.info == infos.pop(0)

    job.signal.register('logging_in', info_cb)
    job.signal.register('logged_in', info_cb)
    job.signal.register('uploading', info_cb)
    job.signal.register('uploaded', info_cb)
    job.signal.register('logging_out', info_cb)
    job.signal.register('logged_out', info_cb)
    await job._submit()
    assert infos == []


@pytest.mark.parametrize('attrname', ('jobs_before_upload', 'jobs_after_upload'))
@pytest.mark.asyncio
async def test_None_and_disabled_are_filtered_from_jobs_attributes(attrname, tmp_path, tracker, tracker_jobs, mocker):
    foo_job = Mock(name='foo', is_enabled=True)
    bar_job = Mock(name='bar', is_enabled=False)
    baz_job = Mock(name='baz', is_enabled=True)
    mocker.patch('upsies.jobs.submit.SubmitJob._submit')
    setattr(
        type(tracker_jobs),
        attrname,
        PropertyMock(return_value=[foo_job, None, bar_job, None, baz_job]),
    )
    job = SubmitJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        tracker=tracker,
        tracker_jobs=tracker_jobs,
    )
    assert getattr(job, attrname) == (foo_job, baz_job)


@pytest.mark.parametrize('submission_ok, exp_hidden', ((False, True), (True, False)))
def test_hidden_property(submission_ok, exp_hidden, job, mocker):
    mocker.patch.object(type(job._tracker_jobs), 'submission_ok', PropertyMock(return_value=submission_ok))
    assert job.hidden is exp_hidden


@pytest.mark.parametrize(
    argnames='submission_ok, jobs, exp_final_job',
    argvalues=(
        (True, (), None),
        (False, (Mock(is_finished=True), Mock(is_finished=False), Mock(is_finished=True)), None),
        (False, (Mock(is_finished=True, id='a'), Mock(is_finished=True, id='b'), Mock(is_finished=True, id='c')), 'c'),
    ),
)
def test_final_job_before_upload_property(submission_ok, jobs, exp_final_job, job, mocker):
    mocker.patch.object(type(job._tracker_jobs), 'submission_ok', PropertyMock(return_value=submission_ok))
    mocker.patch.object(type(job), 'jobs_before_upload', PropertyMock(return_value=jobs))
    if not exp_final_job:
        assert job.final_job_before_upload is None
    else:
        assert job.final_job_before_upload.id == exp_final_job


@pytest.mark.parametrize(
    argnames='own_output, final_job_before_upload, exp_output',
    argvalues=(
        (('torrent url',), None, ('torrent url',)),
        (('torrent url',), Mock(output=('description',)), ('description',)),
    ),
)
def test_output_property(own_output, final_job_before_upload, exp_output, job, mocker):
    parent_class = type(job).mro()[1]
    mocker.patch.object(parent_class, 'output', PropertyMock(return_value=own_output))
    mocker.patch.object(type(job), 'final_job_before_upload', PropertyMock(return_value=final_job_before_upload))
    assert job.output == exp_output


@pytest.mark.parametrize(
    argnames='own_exit_code, final_job_before_upload, exp_exit_code',
    argvalues=(
        (123, None, 123),
        (123, Mock(exit_code=456), 456),
    ),
)
def test_exit_code_property(own_exit_code, final_job_before_upload, exp_exit_code, job, mocker):
    parent_class = type(job).mro()[1]
    mocker.patch.object(parent_class, 'exit_code', PropertyMock(return_value=own_exit_code))
    mocker.patch.object(type(job), 'final_job_before_upload', PropertyMock(return_value=final_job_before_upload))
    assert job.exit_code == exp_exit_code



import functools

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import ConditionalContainer, HorizontalAlign, HSplit, VSplit, Window, WindowAlign
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension

from ....utils import fs, torrent
from .. import widgets
from . import JobWidgetBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class CreateTorrentJobWidget(JobWidgetBase):
    def setup(self):
        self._progress = widgets.ProgressBar()
        self._file_tree = FormattedTextControl()
        self._throughput = FormattedTextControl()
        self._throughput_unit = ''
        self._percent_done = FormattedTextControl()
        self._seconds_elapsed = FormattedTextControl()
        self._seconds_remaining = FormattedTextControl()
        self._seconds_total = FormattedTextControl()
        self._time_started = FormattedTextControl()
        self._time_finished = FormattedTextControl()
        self._current_file_name = ''

        status_column1 = HSplit(
            children=[
                self.make_readout(' Started:', '_time_started', width=8),
                self.make_readout('Finished:', '_time_finished', width=8),
            ],
        )
        status_column2 = HSplit(
            children=[
                self.make_readout('  Elapsed:', '_seconds_elapsed', width=8),
                self.make_readout('Remaining:', '_seconds_remaining', width=8),
                self.make_readout('    Total:', '_seconds_total', width=8),
            ],
        )
        status_column3 = HSplit(
            children=[
                self.make_readout(
                    lambda: '%'.ljust(len(self._throughput_unit)),
                    '_percent_done',
                    width=6,
                    value_align='right',
                    label_side='right',
                ),
                self.make_readout(
                    lambda: self._throughput_unit.ljust(len(self._throughput_unit)),
                    '_throughput',
                    width=6,
                    value_align='right',
                    label_side='right',
                ),
            ],
        )
        status = VSplit(children=[status_column1, status_column2, status_column3], padding=3)

        self._activity_indicator = widgets.ActivityIndicator(
            style='class:info',
        )
        self.job.signal.register('finished', lambda _: self._activity_indicator.disable())

        self._runtime_widget = HSplit(
            children=[
                ConditionalContainer(
                    filter=Condition(lambda: self._progress.percent > 0),
                    content=HSplit([
                        # Estimated times, progress per second, etc
                        status,
                        # Progress bar
                        self._progress,
                    ]),
                ),
                VSplit([
                    # Moving character to indicate activity
                    ConditionalContainer(
                        filter=Condition(lambda: self._activity_indicator.active),
                        content=VSplit([
                            self._activity_indicator,
                            widgets.hspacer,
                        ]),
                    ),
                    # "Hashing", "Searching", "Press ... to skip hashing", etc.
                    Window(
                        FormattedTextControl(self._get_current_activity),
                        # Make sure that the activity is always displayed in
                        # full and the current file name is trimmed to make it fit
                        width=Dimension(weight=10),
                        dont_extend_width=True,
                    ),
                    widgets.hspacer,
                    # Current file name (torrent or content file)
                    ConditionalContainer(
                        filter=Condition(lambda: self._current_file_name),
                        content=Window(
                            FormattedTextControl(lambda: self._current_file_name),
                            style='class:info.readout',
                        ),
                    ),
                ]),
                ConditionalContainer(
                    filter=Condition(lambda: bool(self._file_tree.text)),
                    content=Window(self._file_tree),
                ),
            ],
            style='class:info',
            width=Dimension(min=55, max=150),
        )

        @self.keybindings_global.add(
            'c-t', 'h',
            filter=Condition(lambda: self.job.activity in ('searching', 'verifying')),
        )
        def _(event):
            self.job.cancel_search()

        self.job.signal.register('file_tree', self.handle_file_tree)
        self.job.signal.register('progress_update', self.handle_progress_update)
        self.job.signal.register('finished', lambda _: self.invalidate())

    def make_readout(self, label, attribute, value_align='left', width=None, label_side='left'):
        readout_widget = Window(
            getattr(self, attribute),
            dont_extend_width=True,
            dont_extend_height=True,
            style='class:info.readout',
            width=width,
            align=getattr(WindowAlign, value_align.upper()),
        )
        label_widget = Window(FormattedTextControl(label), dont_extend_width=True)

        if label_side == 'left':
            children = [label_widget, readout_widget]
        elif label_side == 'right':
            children = [readout_widget, label_widget]

        return VSplit(
            children=children,
            padding=1,
            align=HorizontalAlign.CENTER,
        )

    def handle_file_tree(self, file_tree):
        self._file_tree.text = fs.format_file_tree(file_tree)

    def handle_progress_update(self, info):
        self._progress.percent = info.percent_done
        self._percent_done.text = f'{info.percent_done:.2f}'

        self._time_started.text = info.time_started.strftime('%H:%M:%S')
        self._time_finished.text = info.time_finished.strftime('%H:%M:%S')

        def timedelta_string(delta):
            hours, rest = divmod(delta.total_seconds(), 3600)
            minutes, seconds = divmod(rest, 60)
            return f'{hours:02.0f}:{minutes:02.0f}:{seconds:02.0f}'

        self._seconds_elapsed.text = timedelta_string(info.seconds_elapsed)
        self._seconds_remaining.text = timedelta_string(info.seconds_remaining)
        self._seconds_total.text = timedelta_string(info.seconds_total)

        if isinstance(info, torrent.CreateTorrentProgress):
            # Torrent is hashed from fresh files - yummy!
            throughput = info.bytes_per_second.format(
                prefix='binary',
                decimal_places=2,
                trailing_zeros=True,
            )
            self._throughput.text, self._throughput_unit = throughput.split(' ')
            self._throughput_unit += '/s'
            self._current_file_name = fs.basename(info.filepath)

        elif isinstance(info, torrent.FindTorrentProgress):
            # Torrent hashes are searched for in existing torrents
            if info.exception:
                self.job.warn(info.exception)
            elif info.status == 'verifying':
                self._activity_indicator.active = True
            else:
                self._activity_indicator.active = False

            self._throughput.text = f'{info.files_per_second}'
            self._throughput_unit = 'files/s'
            self._current_file_name = fs.basename(info.filepath)

        self.invalidate()

    def _get_current_activity(self):
        if self.job.activity in ('announce_url', 'verifying'):
            self._activity_indicator.active = True
        else:
            self._activity_indicator.active = False

        if self.job.activity == 'announce_url':
            activity = 'Getting announce URL'
        elif self.job.activity == 'searching':
            activity = 'Press <Ctrl-t h> to skip searching for a reusable torrent'
        else:
            activity = self.job.activity.capitalize()

        return activity

    @functools.cached_property
    def runtime_widget(self):
        return self._runtime_widget


class AddTorrentJobWidget(JobWidgetBase):
    def setup(self):
        pass

    @functools.cached_property
    def runtime_widget(self):
        return None


class CopyTorrentJobWidget(JobWidgetBase):
    def setup(self):
        pass

    @functools.cached_property
    def runtime_widget(self):
        return None



from itertools import zip_longest
from unittest.mock import AsyncMock, call

import pytest

from upsies.utils.types import ReleaseType
from upsies.utils.webdbs import Query, SearchResult, tvmaze


@pytest.fixture
def api():
    return tvmaze.TvmazeApi()


def test_sanitize_query(api):
    q = Query('The Foo', type='movie', year='2000')
    assert api.sanitize_query(q) == Query('The Foo', type='unknown', year='2000')
    with pytest.raises(TypeError, match=r'^Not a Query instance: 123$'):
        api.sanitize_query(123)


@pytest.mark.parametrize(
    argnames='text, exp_id',
    argvalues=(
        ('Some Title', None),
        ('1899', None),
        ('myshows/123', None),
        ('shows/123arf', None),
        ('shows/123', '123'),
        ('https://www.tvmaze.com/shows/123', '123'),
        ('https://www.tvmaze.com/shows/123/', '123'),
        ('https://www.tvmaze.com/shows/123/gary-and-his-demons', '123'),
        ('junk shows/123 more junk', '123'),
        ('junk https://www.tvmaze.com/shows/123/stuff more junk', '123'),
    ),
    ids=lambda value: str(value),
)
def test_get_id_from_text(text, exp_id, api, store_response):
    return_value = api.get_id_from_text(text)
    assert return_value == exp_id


@pytest.mark.asyncio
async def test_search_handles_id_in_query(api, store_response):
    results = await api.search(Query(id=35724))
    assert len(results) == 1
    assert results[0].id == 35724
    assert results[0].title == 'Karppi'
    assert results[0].type == ReleaseType.season
    assert results[0].url == 'https://www.tvmaze.com/shows/35724/karppi'
    assert results[0].year == '2018'

    # Make sure we're not awaiting the same coroutine again on subsequent calls
    for _ in range(2):
        assert (await results[0].cast())[:3] == ('Pihla Viitala', 'Lauri Tilkanen', 'Mimosa Willamo')
        assert (await results[0].countries()) == ('Finland',)
        assert (await results[0].directors()) == ()
        assert (await results[0].genres()) == ('drama', 'crime', 'thriller')
        assert (await results[0].summary()) == ('Just months after a tragic loss, detective '
                                                'Sofia Karppi investigates the murder of a '
                                                'woman with ties to a Helsinki construction company.')
        assert (await results[0].title_english()) == 'Deadwind'
        assert (await results[0].title_original()) == 'Karppi'


@pytest.mark.asyncio
async def test_search_returns_empty_list_if_title_is_empty(api, store_response):
    assert await api.search(Query('', year='2009')) == []

@pytest.mark.asyncio
async def test_search_returns_list_of_SearchResults(api, store_response):
    results = await api.search(Query('Star Wars'))
    for result in results:
        assert isinstance(result, SearchResult)


@pytest.mark.parametrize(
    argnames=('query', 'exp_titles'),
    argvalues=(
        (Query('Star Wars', year=2003), ('Star Wars: Clone Wars',)),
        (Query('Star Wars', year='2003'), ('Star Wars: Clone Wars',)),
        (Query('Star Wars', year=2014), ('Star Wars: Rebels',)),
        (Query('Star Wars', year='1990'), ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_year(query, exp_titles, api, store_response):
    results = await api.search(query)
    titles = {r.title for r in results}
    assert titles == set(exp_titles)


@pytest.mark.parametrize(
    argnames=('query', 'exp_titles'),
    argvalues=(
        (Query('star wars clone', type=ReleaseType.series), ('Star Wars: Clone Wars', 'Star Wars: The Clone Wars')),
        (Query('yes minister', type=ReleaseType.series), ('Yes Minister', 'Yes, Prime Minister', 'Yes, Prime Minister')),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_series(query, exp_titles, api, store_response):
    results = await api.search(query)
    titles = {r.title for r in results}
    assert titles == set(exp_titles)


@pytest.mark.parametrize(
    argnames=('query', 'exp_titles'),
    argvalues=(
        (Query('star wars', type=ReleaseType.movie), ()),
        (Query('yes minister', type=ReleaseType.movie), ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_movie(query, exp_titles, api, store_response):
    assert await api.search(query) == []


@pytest.mark.parametrize(
    argnames=('title', 'exp_cast'),
    argvalues=(
        ('Star Wars: Clone Wars', ('André Sogliuzzo', 'John DiMaggio')),
        ('Star Wars: Rebels', ('Taylor Gray', 'Vanessa Marshall')),
        ('Star Wars Resistance', ('Christopher Sean', 'Suzie McGrath')),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_cast(title, exp_cast, api, store_response):
    results = await api.search(Query('Star Wars'))
    results_dict = {r.title: r for r in results}
    cast = await results_dict[title].cast()
    for member in exp_cast:
        assert member in cast


@pytest.mark.parametrize(
    argnames=('title', 'exp_countries'),
    argvalues=(
        ('Star Wars: Clone Wars', ('United States',)),
        ('Something in the Rain', ('South Korea',)),
        ('Le Chalet', ('France',)),
        ('Bron / Broen', ('Sweden',)),
    ),
)
@pytest.mark.asyncio
async def test_search_result_countries(title, exp_countries, api, store_response):
    results = await api.search(Query(title))
    results_dict = {r.title: r for r in results}
    countries = await results_dict[title].countries()
    assert countries == exp_countries


@pytest.mark.parametrize(
    argnames=('title', 'exp_id'),
    argvalues=(
        ('Star Wars: Clone Wars', 1259),
        ('Star Wars: Rebels', 117),
        ('Star Wars Resistance', 36483),
    ),
)
@pytest.mark.asyncio
async def test_search_result_id(title, exp_id, api, store_response):
    results = await api.search(Query('Star Wars'))
    results_dict = {r.title: r for r in results}
    assert results_dict[title].id == exp_id


@pytest.mark.asyncio
async def test_search_result_directors(api, store_response):
    results = await api.search(Query('Star Wars'))
    for result in results:
        directors = await result.directors()
        assert directors == ()


@pytest.mark.parametrize(
    argnames=('title', 'exp_genres'),
    argvalues=(
        ('Star Wars: Clone Wars', ('action', 'science-fiction')),
        ('Something in the Rain', ('drama', 'romance')),
        ('Le Chalet', ('thriller', 'mystery')),
    ),
)
@pytest.mark.asyncio
async def test_search_result_genres(title, exp_genres, api, store_response):
    results = await api.search(Query(title))
    results_dict = {r.title: r for r in results}
    for kw in exp_genres:
        genres = await results_dict[title].genres()
        assert kw in genres


@pytest.mark.parametrize(
    argnames=('title', 'exp_summary'),
    argvalues=(
        ('Star Wars: Clone Wars', 'downfall of the Jedi'),
        ('Star Wars: Rebels', 'set five years before the events of Star Wars: Episode IV'),
        ('Star Wars Resistance', 'Kazuda Xiono'),
    ),
)
@pytest.mark.asyncio
async def test_search_result_summary(title, exp_summary, api, store_response):
    results = await api.search(Query('Star Wars'))
    results_dict = {r.title: r for r in results}
    summary = await results_dict[title].summary()
    assert exp_summary in summary


@pytest.mark.asyncio
async def test_search_result_title_english(api, store_response, mocker):
    mock_title_english = mocker.patch.object(
        api, 'title_english', AsyncMock(return_value='The English Title')
    )
    results = await api.search(Query('Star Wars'))
    for result in results:
        assert await result.title_english() == 'The English Title'
    assert sorted(mock_title_english.call_args_list) == [
        call(result.id) for result in sorted(results, key=lambda r: r.id)
    ]


@pytest.mark.asyncio
async def test_search_result_title_original(api, store_response, mocker):
    mock_title_original = mocker.patch.object(
        api, 'title_original', AsyncMock(return_value='The Original Title')
    )
    results = await api.search(Query('Star Wars'))
    for result in results:
        assert await result.title_original() == 'The Original Title'
    assert sorted(mock_title_original.call_args_list) == [
        call(result.id) for result in sorted(results, key=lambda r: r.id)
    ]


@pytest.mark.parametrize(
    argnames=('id', 'exp_title'),
    argvalues=(
        (1259, 'Star Wars: Clone Wars'),
        (117, 'Star Wars: Rebels'),
        (36483, 'Star Wars Resistance'),
    ),
)
@pytest.mark.asyncio
async def test_search_result_title(id, exp_title, api, store_response):
    results = await api.search(Query('Star Wars'))
    results_dict = {r.id: r for r in results}
    assert results_dict[id].title == exp_title


@pytest.mark.asyncio
async def test_search_result_type(api, store_response):
    results = await api.search(Query('Star Wars'))
    for result in results:
        assert result.type is ReleaseType.series


@pytest.mark.parametrize(
    argnames=('title', 'exp_url'),
    argvalues=(
        ('Star Wars: Clone Wars', 'https://www.tvmaze.com/shows/1259/star-wars-clone-wars'),
        ('Star Wars: Rebels', 'https://www.tvmaze.com/shows/117/star-wars-rebels'),
        ('Star Wars Resistance', 'https://www.tvmaze.com/shows/36483/star-wars-resistance'),
    ),
)
@pytest.mark.asyncio
async def test_search_result_url(title, exp_url, api, store_response):
    results = await api.search(Query('Star Wars'))
    results_dict = {r.title: r for r in results}
    assert results_dict[title].url == exp_url


@pytest.mark.parametrize(
    argnames=('title', 'exp_year'),
    argvalues=(
        ('Star Wars: Clone Wars', '2003'),
        ('Star Wars: Rebels', '2014'),
        ('Star Wars Resistance', '2018'),
    ),
)
@pytest.mark.asyncio
async def test_search_result_year(title, exp_year, api, store_response):
    results = await api.search(Query('Star Wars'))
    results_dict = {r.title: r for r in results}
    assert results_dict[title].year == exp_year


@pytest.mark.parametrize(
    argnames=('id', 'exp_cast'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, (('André Sogliuzzo', 'https://www.tvmaze.com/people/53172/andre-sogliuzzo'),
                ('Anthony Daniels', 'https://www.tvmaze.com/people/49258/anthony-daniels'))),
        # Karppi / Deadwind
        (35724, (('Pihla Viitala', 'https://www.tvmaze.com/people/100790/pihla-viitala'),
                 ('Lauri Tilkanen', 'https://www.tvmaze.com/people/214175/lauri-tilkanen'))),
        # La Foret
        (32614, (('Samuel Labarthe', 'https://www.tvmaze.com/people/83285/samuel-labarthe'),
                 ('Suzanne Clément', 'https://www.tvmaze.com/people/190609/suzanne-clement'))),
        # Terror in the Woods
        (32430, ()),
        # Bron / Broen
        (1910, (('Sofia Helin', 'https://www.tvmaze.com/people/60479/sofia-helin'),
                ('Kim Bodnia', 'https://www.tvmaze.com/people/60480/kim-bodnia'))),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_cast(id, exp_cast, api, store_response):
    cast = (await api.cast(id))[:2]
    if not exp_cast:
        assert cast == ()
    else:
        for person, (name, url) in zip_longest(cast, exp_cast):
            assert str(person) == name
            assert person.url == url

@pytest.mark.asyncio
async def test_cast_deduplicates_actors_with_multiple_roles(api, store_response):
    cast = await api.cast(83)  # The Simpsons
    assert cast.count('Dan Castellaneta') == 1
    assert cast.count('Julie Kavner') == 1
    assert cast.count('Yeardley Smith') == 1
    assert cast.count('Nancy Cartwright') == 1
    assert cast.count('Harry Shearer') == 1


@pytest.mark.parametrize(
    argnames=('id', 'exp_countries'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, ('United States',)),
        # Karppi / Deadwind
        (35724, ('Finland',)),
        # La Foret
        (32614, ('France',)),
        # Terror in the Woods
        (32430, ('United States',)),
        # Bron / Broen
        (1910, ('Sweden',)),
        (None, ()),
    ),
)
@pytest.mark.asyncio
async def test_countries(id, exp_countries, api, store_response):
    countries = await api.countries(id)
    assert countries == exp_countries


@pytest.mark.parametrize(
    argnames=('id', 'exp_creators'),
    argvalues=(
        # Star Wars: Clonse Wars
        # TODO: George Luce is wrong. Maybe it'll get fixed.
        # (1259, (('Genndy Tartakovsky', 'https://www.tvmaze.com/people/56151/genndy-tartakovsky'),)),
        (1259, (('George Lucas', 'https://www.tvmaze.com/people/49259/george-lucas'),)),
        # Karppi / Deadwind
        (35724, (('Rike Jokela', 'https://www.tvmaze.com/people/214183/rike-jokela'),
                 ('Kirsi Porkka', 'https://www.tvmaze.com/people/214184/kirsi-porkka'),
                 ('Jari Olavi Rantala', 'https://www.tvmaze.com/people/214185/jari-olavi-rantala'))),
        # La Foret
        (32614, (('Delinda Jacobs', 'https://www.tvmaze.com/people/79543/delinda-jacobs'),)),
        # Terror in the Woods
        (32430, ()),
        # Bron / Broen
        (1910, (('Hans Rosenfeldt', 'https://www.tvmaze.com/people/76345/hans-rosenfeldt'),
                ('Camilla Ahlgren', 'https://www.tvmaze.com/people/120427/camilla-ahlgren'))),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_creators(id, exp_creators, api, store_response):
    creators = await api.creators(id)
    if not exp_creators:
        assert creators == ()
    else:
        for person, (name, url) in zip_longest(creators, exp_creators):
            assert str(person) == name
            assert person.url == url


@pytest.mark.parametrize(
    argnames=('id', 'exp_directors'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, ()),
        # Karppi / Deadwind
        (35724, ()),
        # La Foret
        (32614, ()),
        # Terror in the Woods
        (32430, ()),
        # Bron / Broen
        (1910, ()),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_directors(id, exp_directors, api, store_response):
    directors = await api.directors(id)
    assert directors == exp_directors


@pytest.mark.parametrize(
    argnames=('id', 'exp_genres'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, ('action', 'family', 'science-fiction')),
        # Karppi / Deadwind
        (35724, ('drama', 'crime', 'thriller')),
        # La Foret
        (32614, ('drama', 'crime')),
        # Terror in the Woods
        (32430, ('mystery', 'supernatural')),
        # Bron / Broen
        (1910, ('drama', 'crime', 'thriller')),
        (None, ()),
    ),
)
@pytest.mark.asyncio
async def test_genres(id, exp_genres, api, store_response):
    genres = await api.genres(id)
    assert genres == exp_genres


@pytest.mark.parametrize(
    argnames='id, season, exp_url',
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, None, 'https://static.tvmaze.com/uploads/images/original_untouched/393/983792.jpg'),
        (1259, 1, 'https://static.tvmaze.com/uploads/images/original_untouched/393/983796.jpg'),
        (1259, 2, 'https://static.tvmaze.com/uploads/images/original_untouched/393/983792.jpg'),
        # Karppi / Deadwind
        (35724, None, 'https://static.tvmaze.com/uploads/images/original_untouched/151/377808.jpg'),
        (35724, 1, 'https://static.tvmaze.com/uploads/images/original_untouched/151/377808.jpg'),
        (35724, 2, 'https://static.tvmaze.com/uploads/images/original_untouched/151/377808.jpg'),
        # La Foret
        (32614, None, 'https://static.tvmaze.com/uploads/images/original_untouched/130/327264.jpg'),
        (32614, 1, 'https://static.tvmaze.com/uploads/images/original_untouched/130/327264.jpg'),
        # Terror in the Woods
        (32430, None, 'https://static.tvmaze.com/uploads/images/original_untouched/356/891645.jpg'),
        (32430, 1, 'https://static.tvmaze.com/uploads/images/original_untouched/240/601573.jpg'),
        (32430, 2, 'https://static.tvmaze.com/uploads/images/original_untouched/356/891645.jpg'),
        # Bron / Broen
        (1910, None, 'https://static.tvmaze.com/uploads/images/original_untouched/11/27700.jpg'),
        (1910, 1, 'https://static.tvmaze.com/uploads/images/original_untouched/252/631183.jpg'),
        (1910, 2, 'https://static.tvmaze.com/uploads/images/original_untouched/252/631184.jpg'),
        (None, None, ''),
    ),
)
@pytest.mark.asyncio
async def test_poster_url(id, season, exp_url, api, store_response):
    poster_url = await api.poster_url(id, season=season)
    assert poster_url == exp_url


@pytest.mark.parametrize(
    argnames=('id', 'exp_rating'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, 8.2),
        # Karppi / Deadwind
        (35724, 6.4),
        # La Foret
        (32614, 6.4),
        # Terror in the Woods
        (32430, 7.7),
        # Bron / Broen
        (1910, 8.5),
        (None, None),
    ),
)
@pytest.mark.asyncio
async def test_rating(id, exp_rating, api, store_response):
    rating = await api.rating(id)
    assert rating == exp_rating


@pytest.mark.parametrize(
    argnames=('id', 'exp_runtimes'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, {'default': 15}),
        # Karppi / Deadwind
        (35724, {'default': 45}),
        # La Foret
        (32614, {'default': 60}),
        # Terror in the Woods
        (32430, {'default': 60}),
        # Bron / Broen
        (1910, {'default': 60}),
        (None, {}),
    ),
)
@pytest.mark.asyncio
async def test_runtimes(id, exp_runtimes, api, store_response):
    runtimes = await api.runtimes(id)
    assert runtimes == exp_runtimes


@pytest.mark.parametrize(
    argnames=('id', 'exp_summary'),
    argvalues=(
        # Star Wars: Clone Wars
        (1259, ('The Clone Wars television series chronicles the events taking place '
                'between Star Wars Episode II: Attack of the Clones and '
                'Star Wars Episode III: Revenge of the Sith. The Clone Wars will ultimately lead '
                'to the downfall of the Jedi and the rise of the Galactic Empire.')),
        # Karppi
        (35724, ('Just months after a tragic loss, detective Sofia Karppi investigates '
                 'the murder of a woman with ties to a Helsinki construction company.')),
        # Le Foret
        (32614, ('A sixteen-year-old girl has been reported missing from her village '
                 'in the Ardennes. With the help of a lonely and mysterious woman, '
                 'Capt. Gaspard Decker and local cop Virginie Musso are tasked to locate her.')),
        # Terror in the Woods
        (32430, ("The great outdoors aren't always so great. In fact, the woods are filled "
                 "with mysteries that sometimes cannot be explained. These are the true accounts "
                 "of people who ventured deep into the forest only to come screaming out with "
                 "stories that defy reality.")),
        # Bron / Broen
        (1910, ('When a body is found on the bridge between Denmark and Sweden, '
                'right on the border, Danish inspector Martin Rohde and Swedish Saga Norén '
                'have to share jurisdiction and work together to find the killer.')),
        (None, ''),
    ),
)
@pytest.mark.asyncio
async def test_summary(id, exp_summary, api, store_response):
    assert await api.summary(id) == exp_summary


@pytest.mark.parametrize(
    argnames=('id', 'exp_title_english', 'exp_title_original'),
    argvalues=(
        (1259, '', 'Star Wars: Clone Wars'),
        (35724, 'Deadwind', 'Karppi'),
        (32614, 'The Forest', 'La Forêt'),
        (32430, 'Terror in the Woods', 'These Woods Are Haunted'),
        (1910, 'The Bridge', 'Bron/Broen'),
        (None, '', ''),
    ),
)
@pytest.mark.asyncio
async def test_title_english_original(id, exp_title_english, exp_title_original, api, store_response):
    assert await api.title_original(id) == exp_title_original
    assert await api.title_english(id) == exp_title_english
    assert await api.title_english(id, default_to_original=False) == exp_title_english
    if exp_title_english:
        assert await api.title_english(id, default_to_original=True) == exp_title_english
    else:
        assert await api.title_english(id, default_to_original=True) == exp_title_original


@pytest.mark.parametrize('id', (1259, 35256, 36072, None))
@pytest.mark.asyncio
async def test_type(id, api, store_response):
    await api.type(id) == ReleaseType.unknown


@pytest.mark.parametrize(
    argnames=('id', 'exp_url'),
    argvalues=(
        (1259, 'https://www.tvmaze.com/shows/1259/star-wars-clone-wars'),
        (35724, 'https://www.tvmaze.com/shows/35724/karppi'),
        (32614, 'https://www.tvmaze.com/shows/32614/la-foret'),
        (32430, 'https://www.tvmaze.com/shows/32430/these-woods-are-haunted'),
        (None, ''),
    ),
)
@pytest.mark.asyncio
async def test_url(id, exp_url, api, store_response):
    assert await api.url(id) == exp_url


@pytest.mark.parametrize(
    argnames=('id', 'exp_year'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, '2003'),
        # Karppi / Deadwind
        (35724, '2018'),
        # La Foret
        (32614, '2017'),
        # Terror in the Woods
        (32430, '2017'),
        # Bron / Broen
        (1910, '2011'),
        (None, ''),
    ),
)
@pytest.mark.asyncio
async def test_year(id, exp_year, api, store_response):
    assert await api.year(id) == exp_year


@pytest.mark.parametrize(
    argnames=('id', 'exp_imdb_id'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, 'tt0361243'),
        # Karppi / Deadwind
        (35724, 'tt6616260'),
        # La Foret
        (32614, 'tt6560040'),
        # Terror in the Woods
        (32430, 'tt7534328'),
        # Bron / Broen
        (1910, 'tt1733785'),
        (None, ''),
    ),
)
@pytest.mark.asyncio
async def test_imdb_id(id, exp_imdb_id, api, store_response):
    assert await api.imdb_id(id) == exp_imdb_id


@pytest.mark.parametrize(
    argnames=('id', 'season', 'episode', 'exp_episode'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, 1, 5, {
            'date': '2003-11-13',
            'episode': '5',
            'season': '1',
            'summary': ('On Mon Calamari, Kit Fisto and his Scuba Troopers defends the '
                        'Calamari council against Manta Droid sub fighters army of the '
                        'Quarren Isolation league. Mon Calamari Knights riding giant '
                        "Keelkana's provide the Republic forces with back up."),
            'title': 'Chapter V',
            'url': 'https://www.tvmaze.com/episodes/109474/star-wars-clone-wars-1x05-chapter-v',
        }),

        # Karppi / Deadwind
        (35724, 3, 4, {
            'date': '2021-10-29',
            'episode': '4',
            'season': '3',
            'summary': ('Laura is the target of an assassination attempt and the '
                        'management of the pharmaceutical company is entangled in a '
                        'tangle. Karppi and Nurmi find their way to the hideout used by '
                        'the murderer and at the same time access the traces of the symbol '
                        'seen in connection with the bodies. The dual role takes Henna '
                        'deeper into the underworld.'),
            'title': 'Kos',
            'url': 'https://www.tvmaze.com/episodes/2204669/karppi-3x04-kos',
        }),

        # La Foret
        (32614, 1, 2, {
            'date': '2017-05-30',
            'episode': '2',
            'season': '1',
            'summary': ('Capt. Decker questions Maya, stoking tensions at the station. A '
                        'grim discovery in the woods and two more disappearances put the '
                        'town on edge.'),
            'title': 'Épisode 2',
            'url': 'https://www.tvmaze.com/episodes/1329360/la-foret-1x02-episode-2',
        }),

        (None, 1, 3, {
            'date': '',
            'episode': '',
            'season': '',
            'summary': '',
            'title': '',
            'url': '',
        }),
    ),
)
@pytest.mark.asyncio
async def test_episode(id, season, episode, exp_episode, api, store_response):
    assert await api.episode(id, season, episode) == exp_episode


@pytest.mark.parametrize(
    argnames=('id', 'exp_status'),
    argvalues=(
        # Star Wars: Clonse Wars
        (1259, 'Ended'),
        # Karppi / Deadwind
        (35724, 'Ended'),
        # La Foret
        (32614, 'Ended'),
        # Terror in the Woods
        (32430, 'Running'),
        # Bron / Broen
        (1910, 'Ended'),
        (None, ''),
    ),
)
@pytest.mark.asyncio
async def test_status(id, exp_status, api, store_response):
    assert await api.status(id) == exp_status
